(*Title: View-Based Owicki-Gries Reasoning for Persistent x86-TSO 
  Authors: Eleni Bila, Brijesh Dongol, Ori Lahav, Azalea Raad, John Wickerson*)

theory CAS_Rules
  imports "../State" "../Language"  "../Assertions"  "../Wellformedness"
begin

(* The lemmas that begin with the  comment "rule [name]: __" are the proof rules regarding the CAS instruction. The
name corresponds to the name of the rule in the paper. 
   All the other lemmas are auxiliary and are used for proving the proof rules.
   Rules that end with "lc" concern mostly local correctness.
   Rules that end with "ni" concern  mostly  non-interference.
   Rules that end with "gen" concern non-interference and local correctness.
   Rules that end with "rel" concern relations between view sets. *) 

(*rule: If the value of the last write on x is different from v1 the CAS fails and so r = 0.*)
lemma cas_nlv_lc:
  assumes "mem_structured ts"
 and "vbounded ts"
and " step t ( CAS x  v1 v2 r) ts ts' "
and "   \<lceil>x:v3\<rceil> ts "
and "v1 \<noteq> v3"
shows "     regs ts' t r = 0 "
  using assms
  apply (simp add: step_def)
  apply (simp add:  cas_fail_trans_def)
  by force

lemma cas_suc_reg:
assumes " ts' =   cas_succ_trans t ind x v1 v2 r ts "
 and "vbounded ts"
 and "mem_structured ts"
shows "  regs ts' t r = 1 "
  using assms
  by (simp add:  cas_succ_trans_def)


lemma cas_fail_reg:
assumes " ts' =   cas_fail_trans t ind x v1 v2 r ts "
 and "vbounded ts"
 and "mem_structured ts"
shows "  regs ts' t r = 0 "
  using assms
  by (simp add:  cas_fail_trans_def)


lemma cas_succ_lastentry:
  assumes "mem_structured ts"
 and "vbounded ts"
 and " ts' =   cas_succ_trans t ind x v1 v2 r ts"
shows  " last_entry ts' x = length(memory ts')-1"
  using assms
  apply (subgoal_tac "  comploc ((memory ts') !(length(memory ts')-1)) x = x")
  prefer 2
    apply (simp add: store_trans_def)
   apply (subgoal_tac "( length(memory ts')-1 )\<in> last_entry_set ts' x")
    prefer 2
    apply (simp add: last_entry_set_def)
   apply (subgoal_tac " \<forall>i. (i \<in> last_entry_set ts' x) \<longrightarrow> i \<le> (length(memory ts')-1)")
    prefer 2
    apply (simp add: last_entry_set_def)
   apply (simp add: last_entry_def )
   apply (subgoal_tac" finite (last_entry_set ts' x )")
    prefer 2
  apply (simp add: last_entry_set_def)
  using Max_eqI by auto[1]


(*rule: If the CAS succeeds then the value of the last write on x becomes v2.*)
lemma cas_succ_lv_lc:
  assumes "mem_structured ts"
 and "vbounded ts"
and " step t ( CAS x  v1 v2 r) ts ts' "
and "    regs ts' t r = 1 "
shows "     \<lceil>x:v2\<rceil> ts' "
  using assms
  apply (simp add: step_def)
  apply clarify
  apply (subgoal_tac " ts' =   cas_fail_trans t ind x v1 v2 r ts \<longrightarrow>  regs ts' t r = 0 ")
   prefer 2
   apply (simp add: cas_fail_trans_def)
  apply (subgoal_tac " ts' \<noteq>    cas_fail_trans t ind x v1 v2 r ts" )
   prefer 2
   apply linarith
  apply (subgoal_tac " \<lceil>x:v1\<rceil> ts \<and>  ind = last_entry ts x")
   prefer 2
   apply (metis compval_def)
  apply (subgoal_tac  " ts' =    cas_succ_trans t ind x v1 v2 r ts")
   prefer 2
   apply presburger
  apply (subgoal_tac " last_entry ts' x = length(memory ts')-1") 
   prefer 2
   apply (meson cas_succ_lastentry)
  by simp

lemma [simp]:  " ts' =   cas_succ_trans t ind x v1 v2 r ts \<longrightarrow>  regs ts' t r = 1 "
  by (simp add: cas_succ_trans_def)

lemma [simp]: " ts' =   cas_fail_trans t ind x v1 v2 r ts \<longrightarrow>  regs ts' t r = 0 "
  by (simp add: cas_fail_trans_def)



(*rule [CP1]: After thread t executes CAS, either r =1 and the value of last write on x becomes v1, or r=0.*)
lemma cas_lc:
  assumes "mem_structured ts"
 and "vbounded ts"
and " step t ( CAS x  v1 v2 r) ts ts' "
shows "    ( \<lceil>x:v2\<rceil> ts' \<and>  regs ts' t r = 1 ) \<or>  regs ts' t r= 0   "
  using assms
 apply (simp add: step_def)
  apply clarify
  apply (subgoal_tac " regs ts' t r = 1 \<longrightarrow> ts' = cas_succ_trans t ind x v1 v2 r ts")
   prefer 2
   apply auto[1]
  apply  (subgoal_tac " regs ts' t r = 1 \<longrightarrow>   \<lceil>x:v2\<rceil> ts'")
  prefer 2
  using assms(3) cas_succ_lv_lc apply blast
 apply (subgoal_tac " regs ts' t r = 0 \<longrightarrow> ts' = cas_fail_trans t ind x v1 v2 r ts")
   prefer 2
   apply auto[1]
  by (metis One_nat_def cas_fail_reg cas_suc_reg compval_def)



lemma cas_ots_lastentry:
 assumes  "mem_structured ts"
and "  ts' = cas_succ_trans ti ind x v1 v2 r ts "
shows " OTS ts' ti x = { length(memory ts')-1} "
using assms 
  apply(simp add:  mem_structured_def   OTS_def OTSF_def cas_succ_trans_def notoverwritten_def
  )
 apply (case_tac " 0 \<notin> OTS ts' ti x" )
   prefer 2
  using  assms   apply (simp add: OTS_def OTSF_def notoverwritten_def )
  using assms   apply  (subgoal_tac" ( coh ts' ti x > 0 )" )
    prefer 2
apply simp
    apply clarsimp
  by(auto simp add: split: Msg.split)



lemma cas_succ_ov_lc:
 assumes "mem_structured ts"
 and "vbounded ts"
 and " ts' = cas_succ_trans ti ind x v1 v2 r ts "
shows "  [x]\<^sub>ti ts' = {v2}"
  using assms 
  apply (subgoal_tac " OTS ts' ti x = { length(memory ts')-1}")
   prefer 2
  using cas_ots_lastentry apply blast
  by (simp add: mapval_def)


lemma [simp]:
  assumes "mem_structured ts"
  and " i \<in>  OTS ts ti x"
shows " 0 \<le> i  \<and> i < length(memory ts) "
  using assms apply (simp add:  mem_structured_def OTS_def OTSF_def )
  done


lemma [simp]:
  assumes "mem_structured ts"
 and " 0 \<le> i  \<and> i < length(memory ts) "
 and "memory ts ! i = Init_Msg"
shows "i = 0"
  using assms(1)  apply ( unfold  mem_structured_def   )
  apply (subgoal_tac"  length (memory ts) \<noteq>  0 " )
   prefer 2
   apply blast
  using assms(2) assms(3) le_neq_implies_less by blast



lemma cas_succ_ots_daddr_st:
assumes "mem_structured ts"
and "vbounded ts"
and " ts' = cas_succ_trans t' ind lx v1 v2 r ts "
and " lx \<noteq> y "
shows " OTS ts' t y \<subseteq> OTS ts t y"
  using assms
 apply (subgoal_tac "  coh ts'  t y =  coh ts  t y")
    prefer 2
  apply (simp add: cas_succ_trans_def)

 apply (subgoal_tac "  vrnew ts  t \<le>  vrnew ts'  t ")
  prefer 2
     apply (simp add: cas_succ_trans_def)
    apply (simp add: OTS_def)
    apply (simp add: OTSF_def)
    apply (simp add: notoverwritten_def)
    apply safe
                  apply (metis Msg.sel(1) less_antisym nth_append_length)
  apply (metis bot.extremum_strict bot.extremum_unique bot_nat_def less_SucE mem_l mem_structured_def not_less_iff_gr_or_eq store_add)
                apply (metis Msg.sel(1) not_less_less_Suc_eq nth_append_length)
  apply (metis Msg.sel(1) bot.extremum bot_nat_def less_SucE mem_l nth_append_length store_add)
  apply (metis bot_nat_0.extremum le_trans less_eq_Suc_le mem_l not_le_imp_less not_less_iff_gr_or_eq store_add)
             apply (metis Msg.distinct(1) less_antisym nth_append_length)
  apply (metis dual_order.trans le0 less_SucI mem_l store_add)
           apply (metis Msg.distinct(1) less_SucE nth_append_length)
  apply (metis Msg.distinct(1) less_SucE less_nat_zero_code mem_l not_le_imp_less nth_append_length store_add)
  apply (metis bot.extremum bot_nat_def length_append_singleton mem_structured_def store_add store_wfs verit_comp_simplify1(3) verit_la_disequality)
        apply (metis Msg.sel(1) less_antisym nth_append_length)
  apply (metis bot_nat_0.extremum le_trans less_eq_Suc_le mem_l not_le_imp_less not_less_iff_gr_or_eq store_add)
  apply (metis Msg.sel(1) less_antisym nth_append_length)
     apply (metis Msg.sel(1) bot.extremum bot_nat_def less_SucE mem_l nth_append_length store_add)
  by (metis (no_types) bot.extremum bot_nat_def dual_order.strict_trans le_trans less_not_refl mem_l not_less_less_Suc_eq store_add)


lemma cas_succ_ov_daddr_st:
assumes "mem_structured ts"
and "vbounded ts"
and " ts' = cas_succ_trans t' ind lx v1 v2 r ts "
and " lx \<noteq> y "
shows " [y]\<^sub>t ts'  \<subseteq> [y]\<^sub>t ts"
  using assms
  apply (subgoal_tac " OTS ts' t y \<subseteq> OTS ts t y ")
   prefer 2
   apply (simp add: cas_succ_ots_daddr_st)
 apply (subgoal_tac " mapval (OTS ts t y)  (memory ts') = mapval (OTS ts t y) (memory ts )")
 prefer 2
    apply (subgoal_tac "\<forall> i. i \<in> (OTS ts t y) \<longrightarrow> i < length (memory ts)")
    prefer 2
    apply (simp add: OTS_def)
  apply (unfold mapval_def )
  using mem_l apply auto[1]
  by blast


lemma cas_succ_ots_daddr_dt:
assumes "mem_structured ts"
and "vbounded ts"
and " ts' = cas_succ_trans t' ind lx v1 v2 r ts "
and " lx \<noteq> y "
and "t \<noteq> t' "
shows " OTS ts' t y =  OTS ts t y"
  using assms
 apply (subgoal_tac "  coh ts'  t y =  coh ts  t y")
    prefer 2
  apply (simp add: cas_succ_trans_def)
 apply (subgoal_tac "  vrnew ts  t =  vrnew ts'  t ")
  prefer 2
    apply (simp add: cas_succ_trans_def)
   apply (subgoal_tac " (\<forall>i.(i\<ge>0\<and>i<length(memory ts)) \<longrightarrow> ( memory ts)!i =( memory ts')!i) ")
    prefer 2
  using mem_l_casucc apply blast
   apply (subgoal_tac" length( memory ts) \<notin>  OTS ts' t y")
  prefer 2
 apply (simp add: OTS_def)
    apply (simp add: OTSF_def)
  apply (simp add: OTS_def)
   apply (simp add: OTSF_def)
  apply (simp add: notoverwritten_def)
  apply safe
  apply (simp_all)
                      apply (metis Msg.sel(1) not_less_less_Suc_eq nth_append_length)                 
                      apply (metis Msg.sel(1) not_less_less_Suc_eq nth_append_length)
  apply (metis Msg.sel(1) less_Suc_eq nth_append_length)                 
                      apply (metis Msg.discI not_less_less_Suc_eq nth_append_length)
  using less_Suc_eq apply blast
                      apply (metis Msg.distinct(1) less_antisym nth_append_length)
                      apply (metis Msg.discI less_Suc_eq nth_append_length)
                     apply (metis Msg.sel(1) not_less_less_Suc_eq nth_append_length)             
                   apply (metis Msg.sel(1) not_less_less_Suc_eq nth_append_length)
                  apply (metis Msg.sel(1) less_Suc_eq nth_append_length)  
            apply (metis Msg.sel(1) less_Suc_eq nth_append_length)       
          apply (metis Msg.sel(1) less_Suc_eq nth_append_length) 
     apply (metis Msg.sel(1) less_Suc_eq nth_append_length)
  by (metis Msg.sel(1) less_Suc_eq nth_append_length)


lemma cas_succ_ov_daddr_dt:
assumes "mem_structured ts"
and "vbounded ts"
and " ts' = cas_succ_trans t' ind lx v1 v2 r ts "
and " lx \<noteq> y "
and " t \<noteq> t'"
shows " [y]\<^sub>t ts'  = [y]\<^sub>t ts"
  using assms
  apply (subgoal_tac "OTS ts' t y =  OTS ts t y ")
   prefer 2
  using cas_succ_ots_daddr_dt apply blast
   apply (subgoal_tac " mapval (OTS ts t y)  (memory ts') = mapval (OTS ts t y) (memory ts )")
 prefer 2
    apply (subgoal_tac "\<forall> i. i \<in> (OTS ts t y) \<longrightarrow> i < length (memory ts)")
    prefer 2
    apply (simp add: OTS_def)
  apply (unfold mapval_def )
  using mem_l apply auto[1]
  by blast


lemma cas_ots_dt_lc:
  assumes "mem_structured ts"
and "vbounded ts"
 and " ts' = cas_succ_trans t' ind x v1 v2 r ts "
and " t \<noteq> t'"
shows " OTS ts' t x  = OTS ts t x \<union>  { length(memory ts')-1} "
 using assms
  apply (subgoal_tac "  coh ts t  x = coh ts' t x")
  prefer 2
  apply (simp add: cas_succ_trans_def)
  apply (subgoal_tac "  vrnew ts t  = vrnew  ts' t ")
   prefer 2
  apply (simp add: cas_succ_trans_def)
  apply (subgoal_tac "  memory ts'  =  memory ts  @ [msg x v2 t'] ")
   prefer 2
   apply auto[1] 
  apply (simp add: OTS_def)
  apply (simp add: OTSF_def)
  apply safe
                     apply (simp_all)
  apply (metis bot.extremum_unique bot.not_eq_extremum bot_nat_def less_SucE mem_l mem_structured_def store_add)
                   apply (simp add: mem_l)
                  apply (simp add: notoverwritten_def)
  apply (metis bot_nat_0.extremum_uniqueI le_imp_less_Suc less_imp_le_nat mem_l nat_le_linear store_add)
                 apply (simp add: notoverwritten_def)
  apply (metis bot_nat_0.extremum_uniqueI le_imp_less_Suc less_imp_le_nat mem_l nat_le_linear store_add)
                apply (metis bot_nat_0.extremum_uniqueI less_SucE mem_l nat_le_linear store_add)
  apply (metis bot_nat_0.extremum_uniqueI less_SucE mem_l nat_le_linear store_add)
              apply (simp add: notoverwritten_def)
  apply (metis bot_nat_0.extremum_uniqueI le_imp_less_Suc less_imp_le_nat mem_l nat_le_linear store_add)
             apply (simp add: mem_l)
  apply (simp add: notoverwritten_def)
  apply (metis bot_nat_0.extremum_uniqueI le_imp_less_Suc less_imp_le_nat mem_l nat_le_linear store_add)
           apply(simp add: vbounded_def)     
           apply (metis nat_less_le)
          apply(simp add: vbounded_def) 
          apply (simp add: notoverwritten_def)
         apply (simp add: notoverwritten_def)
  apply (metis bot.extremum bot_nat_def mem_l store_add)
  apply (metis bot.extremum bot_nat_def mem_l store_add)

     apply (simp add: notoverwritten_def)
 apply (subgoal_tac " vrnew ts t <  (length (memory ts))")
        prefer 2
        apply(simp add: vbounded_def)
        apply metis
apply (subgoal_tac " vrnew ts t <  Suc(length (memory ts))")
        prefer 2
        apply linarith
       apply simp
       apply (subgoal_tac " \<forall>i.  i \<le> vrnew ts t \<and>  xa < i \<longrightarrow>   loc (memory ts ! i) \<noteq> loc (memory ts ! xa)")
        prefer 2
        apply auto[1]
       apply (simp add: nth_append)
      apply (simp add: notoverwritten_def)
 apply (subgoal_tac " vrnew ts t <  (length (memory ts))")
        prefer 2
        apply(simp add: vbounded_def)
        apply metis
apply (subgoal_tac " vrnew ts t <  Suc(length (memory ts))")
        prefer 2
       apply linarith
      apply (simp add: nth_append)
 apply (simp add: notoverwritten_def)
 apply (subgoal_tac " vrnew ts t <  (length (memory ts))")
        prefer 2
        apply(simp add: vbounded_def)
      apply metis
     apply (simp add: nth_append)
  apply (simp add: notoverwritten_def)
 apply (subgoal_tac " vrnew ts t <  (length (memory ts))")
        prefer 2
        apply(simp add: vbounded_def)
     apply metis
    apply (simp add: nth_append)
 apply (simp add: notoverwritten_def)
 apply (subgoal_tac " vrnew ts t <  (length (memory ts))")
        prefer 2
        apply(simp add: vbounded_def)
      apply metis
     apply (simp add: nth_append)
  apply (simp add: notoverwritten_def)
 apply (subgoal_tac " vrnew ts t <  (length (memory ts))")
        prefer 2
        apply(simp add: vbounded_def)
     apply metis
  apply (simp add: nth_append)
 apply (simp add: notoverwritten_def)
 apply (subgoal_tac " vrnew ts t <  (length (memory ts))")
        prefer 2
        apply(simp add: vbounded_def)
      apply metis
  by (simp add: nth_append)


lemma  cas_succ_ov_dt_lc:
  assumes "mem_structured ts"
  and "vbounded ts"
  and "t \<noteq> t'"
 and " ts' = cas_succ_trans t' ind x v1 v2 r ts "
shows "  [x]\<^sub>t ts'  =  [x]\<^sub>t ts  \<union> {v2}"
  using assms
   apply (subgoal_tac  " OTS ts' t x  = OTS ts t x \<union>  { length(memory ts')-1} ")
  prefer 2
   apply (simp add: cas_ots_dt_lc)
  apply (subgoal_tac " mapval (OTS ts t x)  (memory ts') = mapval (OTS ts t x) (memory ts )")
 prefer 2
    apply (subgoal_tac "\<forall> i. i \<in> (OTS ts t x) \<longrightarrow> i < length (memory ts)")
    prefer 2
    apply (simp add: OTS_def)
  apply (unfold mapval_def )
  using mem_l apply auto[1]
  by simp


lemma cas_fail_ov_ni:
  assumes "mem_structured ts"
  and "vbounded ts"
 and "t \<noteq> t'"
 and " ts' = cas_fail_trans t' ind x v1 v2 r ts "
shows "    [y]\<^sub>t ts  =   [y]\<^sub>t ts'"
  using assms 
apply (subgoal_tac " OTS ts t y = OTS ts' t y")
  prefer 2
  apply (case_tac " x= y")
    apply (simp add: OTS_def  )
    apply (subgoal_tac " vrnew ts t = vrnew ts' t")
     prefer 2
     apply (simp add: cas_fail_trans_def)
apply (simp add: OTSF_def)
    apply (subgoal_tac "memory ts = memory ts'")
    prefer 2
  apply simp
  apply (subgoal_tac " coh ts t y =  coh ts' t y")
  prefer 2
      apply (simp add: cas_fail_trans_def)
    apply (smt Collect_cong notoverwritten_def)
   apply (simp add: OTS_def)
 apply (subgoal_tac " vrnew ts t = vrnew ts' t")
     prefer 2
      apply (simp add: cas_fail_trans_def)
apply (simp add: OTSF_def)
    apply (subgoal_tac "memory ts = memory ts'")
    prefer 2
  apply simp
  apply (subgoal_tac " coh ts t y =  coh ts' t y")
  prefer 2
     apply (simp add: cas_fail_trans_def)
  apply (smt Collect_cong notoverwritten_def)
  by (simp add: mapval_def)


lemma coh_cas_fail_st_dadrr :
  assumes "mem_structured ts"
  and "vbounded ts"
  and " ts' = cas_fail_trans t  ind x v1 v2 r ts "
   and " x \<noteq> y"
 shows " coh ts t y  = coh ts' t y"
  using assms 
  by (simp add:  cas_fail_trans_def Let_def)

(*rule [CP2]: After thread t executes a CAS on y,the observable values of any other address for t might become less.*)
lemma cas_ov_daddr_ni:
  assumes "mem_structured ts"
 and "vbounded ts"
and " step t' ( CAS y  v1 v2 r) ts ts' "
and " x \<noteq> y"
shows " [x]\<^sub>t ts' \<subseteq> [x]\<^sub>t ts "
 using assms
 apply (simp add: step_def)
  apply clarify
  apply (case_tac "  ts' = cas_succ_trans t' ind y v1 v2 r ts")
  apply simp
 apply (case_tac "t = t'")
    prefer 2
    apply (subgoal_tac " [x]\<^sub>t ts' = [x]\<^sub>t ts")
     prefer 2
  using cas_succ_ov_daddr_dt apply presburger
    apply (simp add: mapval_def)
  apply (metis (no_types, lifting) UnE UnI1 UnI2)
   apply (subgoal_tac " [x]\<^sub>t ts' \<subseteq> [x]\<^sub>t ts")
    prefer 2
  using cas_succ_ov_daddr_st apply presburger
  apply (simp add: mapval_def)
   apply (meson UnE in_mono)
 apply (case_tac "t = t'")
   prefer 2
  using cas_fail_ov_ni apply blast

 apply (subgoal_tac" OTS ts' t x \<subseteq> OTS ts t x")
   prefer 2
apply (subgoal_tac " coh ts t x \<le> coh ts' t x")
    prefer 2
    apply (simp add: cas_fail_trans_def Let_def)
 apply clarify
  apply (case_tac "ind = coh ts t x")
   apply simp
  apply (simp add: OTS_def)
    apply (simp add: OTSF_def)
 apply (subgoal_tac "  vrnew ts t   \<le> vrnew  ts' t ")
    prefer 2
     apply (simp add:  cas_fail_trans_def Let_def)
    apply (elim conjE)
    apply (simp add: notoverwritten_def)
apply (subgoal_tac "  vrnew ts t   \<le> vrnew  ts' t ")
    prefer 2
    apply (simp add:  cas_fail_trans_def Let_def)
apply (simp add: OTS_def )
   apply (simp add: OTSF_def notoverwritten_def )
  apply (subgoal_tac "  memory ts   =  memory  ts' ")
    prefer 2
   apply simp
  apply (simp add: mapval_def)
  by blast

(*rule: cas_ov_daddr_ni for singleton OV set*)
corollary  cas_ov_daddr_ni_s:
  assumes "mem_structured ts"
 and "vbounded ts"
and " step t' ( CAS y  v1 v2 r) ts ts' "
and " x \<noteq> y"
and " [x]\<^sub>t ts =  {w}"
and "  (\<forall> t addr.  [addr]\<^sub>t ts \<noteq> {} )"
and  "\<forall>ti l. comploc ( (memory ts)!(coh ts ti l)) l = l"
shows " [x]\<^sub>t ts' = [x]\<^sub>t ts "
  using assms
  by (metis OV_notempty_preserved cas_ov_daddr_ni subset_singleton_iff)


lemma cas_nov_daddr_ni:
  assumes "mem_structured ts"
 and "vbounded ts"
and " step t' ( CAS y  v1 v2 r) ts ts' "
and " v3 \<notin>[x]\<^sub>t ts "
and " x \<noteq> y"
shows "v3 \<notin>[x]\<^sub>t ts' "
  using assms
 apply (simp add: step_def)
  apply clarify
  apply (case_tac "  ts' = cas_succ_trans t' ind y v1 v2 r ts")
  apply simp
 apply (case_tac "t = t'")
    prefer 2
    apply (subgoal_tac " [x]\<^sub>t ts' = [x]\<^sub>t ts")
     prefer 2
  apply (metis assms(5) cas_succ_ov_daddr_dt)
    apply (simp add: mapval_def)
  apply (metis (no_types, lifting) UnE UnI1 UnI2)
   apply (subgoal_tac " [x]\<^sub>t ts' \<subseteq> [x]\<^sub>t ts")
    prefer 2
  using cas_succ_ov_daddr_st apply presburger
  apply (simp add: mapval_def)
   apply (meson UnE in_mono)
 apply (case_tac "t = t'")
   prefer 2
  using cas_fail_ov_ni apply blast

 apply (subgoal_tac" OTS ts' t x \<subseteq> OTS ts t x")
   prefer 2
apply (subgoal_tac " coh ts t x \<le> coh ts' t x")
    prefer 2
    apply (simp add: cas_fail_trans_def Let_def)
 apply clarify
  apply (case_tac "ind = coh ts t x")
   apply simp
  apply (simp add: OTS_def)
    apply (simp add: OTSF_def)
 apply (subgoal_tac "  vrnew ts t   \<le> vrnew  ts' t ")
    prefer 2
     apply (simp add:  cas_fail_trans_def Let_def)
    apply (elim conjE)
    apply (simp add: notoverwritten_def)
apply (subgoal_tac "  vrnew ts t   \<le> vrnew  ts' t ")
    prefer 2
    apply (simp add:  cas_fail_trans_def Let_def)
apply (simp add: OTS_def )
   apply (simp add: OTSF_def notoverwritten_def )
  apply (subgoal_tac "  memory ts   =  memory  ts' ")
    prefer 2
   apply simp
 apply (subgoal_tac " \<forall>i.  i \<notin> OTS ts t' x \<longrightarrow>  i \<notin> OTS ts' t' x ")
  prefer 2
  apply blast
  apply (unfold mapval_def)

  by (metis (mono_tags, lifting) image_iff)


lemma cas_nov_lc:
  assumes "mem_structured ts"
 and "vbounded ts"
and " step t' ( CAS x  v1 v2 r) ts ts' "
and " v3 \<notin>[x]\<^sub>t ts "
and "v2 \<noteq> v3"
shows "v3 \<notin>[x]\<^sub>t ts' "
  using assms
  apply (simp add: step_def)
  apply clarify
  apply (case_tac "  ts' = cas_succ_trans t' ind x v1 v2 r ts")
  apply (metis UnE cas_succ_ov_dt_lc cas_succ_ov_lc singletonD)
  apply simp
 apply (case_tac "t = t'")
   prefer 2
   apply (subgoal_tac " [x]\<^sub>t ts = [x]\<^sub>t ts'")
    prefer 2
  using cas_fail_ov_ni apply blast
   apply (subgoal_tac "  v3 \<notin>[x]\<^sub>t ts'")
    prefer 2
    apply blast
   apply simp
 apply (subgoal_tac" OTS ts' t' x \<subseteq> OTS ts t' x")
   prefer 2
apply (subgoal_tac " coh ts t' x \<le> coh ts' t' x")
    prefer 2
    apply (simp add: cas_fail_trans_def Let_def)
 apply clarify
  apply (case_tac "ind = coh ts t' x")
   apply simp
  apply (simp add: OTS_def)
    apply (simp add: OTSF_def)
 apply (subgoal_tac "  vrnew ts t'   \<le> vrnew  ts' t' ")
    prefer 2
    apply (simp add:  cas_fail_trans_def Let_def)
 apply (simp add: OTS_def )
   apply (simp add: OTSF_def notoverwritten_def )
  apply (subgoal_tac "  memory ts   =  memory  ts' ")
    prefer 2
    apply simp
 apply clarify
   apply (intro conjI impI)
         apply presburger
        apply (meson dual_order.trans)
       apply (metis dual_order.trans)
      apply blast
     apply blast
    apply (meson dual_order.trans)
   apply (meson dual_order.trans)
  apply (subgoal_tac " \<forall>i.  i \<notin> OTS ts t' x \<longrightarrow>  i \<notin> OTS ts' t' x ")
  prefer 2
   apply auto[1]
  apply (simp add: mapval_def)
  by auto

(*rule [CS2]: If the  value of the last write on x is v3 and a thread t executes CAS x  v1 v2 r, and v3 different to v1, the value
of the last write on x remains the same.*)
lemma cas_fail_lv_ni:
  assumes "mem_structured ts"
 and "vbounded ts"
and "  \<lceil>x:v3\<rceil> ts "
and " v3 \<noteq> v1 "
and " step t ( CAS x  v1 v2 r) ts ts' "
shows "     \<lceil>x:v3\<rceil> ts' "
  using assms
  apply (simp add: step_def)
  apply clarify
  apply (subgoal_tac " ts' = cas_fail_trans t ind x v1 v2 r ts")
   prefer 2
   apply simp
  apply simp
  apply (subgoal_tac " memory ts = memory ts'")
   prefer 2
   apply simp
  apply (simp(no_asm) add: last_entry_def)
  by (simp(no_asm) add: last_entry_set_def)


(*rule [CS1]: If the  value of the last write on x is v3 and a thread t' executes CAS x  v1 v2 r, and v3 different to v1, and v4 is
not in the observable values of x for thread t, then after the execution of the CAS, v4 is still not in  the observable values of x for thread t'.*)
lemma cas_fail_nov_ni:
  assumes "mem_structured ts"
 and "vbounded ts"
and "  \<lceil>x:v3\<rceil> ts "
and " v4 \<notin>[x]\<^sub>t ts "
and " v3 \<noteq> v1 "
and " step t' ( CAS x  v1 v2 r) ts ts' "
shows  " v4 \<notin>[x]\<^sub>t ts' "
  using assms
 apply (simp add: step_def)
  apply clarify
  apply (subgoal_tac " ts' = cas_fail_trans t ind x v1 v2 r ts")
   prefer 2
  apply (metis cas_fail_ov_ni)
   apply simp
 apply (case_tac "t = t'")
   prefer 2
   apply (subgoal_tac " [x]\<^sub>t ts = [x]\<^sub>t ts'")
    prefer 2
  using cas_fail_ov_ni apply blast
   apply (subgoal_tac "  v4 \<notin>[x]\<^sub>t ts'")
    prefer 2
    apply blast
   apply simp
 apply (subgoal_tac" OTS ts' t' x \<subseteq> OTS ts t' x")
   prefer 2
apply (subgoal_tac " coh ts t' x \<le> coh ts' t' x")
    prefer 2
    apply (simp add: cas_fail_trans_def Let_def)
  apply (case_tac "ind = coh ts t' x")
   apply simp
  apply (simp add: OTS_def)
    apply (simp add: OTSF_def)
 apply (subgoal_tac "  vrnew ts t'   \<le> vrnew  ts' t' ")
    prefer 2
    apply (simp add:  cas_fail_trans_def Let_def)
 apply (simp add: OTS_def )
   apply (simp add: OTSF_def notoverwritten_def )
  apply (subgoal_tac "  memory ts   =  memory  ts' ")
    prefer 2
     apply simp
    apply clarify
   apply (intro conjI impI)
         apply presburger
        apply (meson dual_order.trans)
       apply (metis dual_order.trans)
      apply blast
     apply blast
    apply (meson dual_order.trans)
   apply (meson dual_order.trans)
  apply (subgoal_tac " \<forall>i.  i \<notin> OTS ts t' x \<longrightarrow>  i \<notin> OTS ts' t' x ")
  prefer 2
   apply auto[1]
  apply (simp add: mapval_def)
  by auto


(*rule [CP3]: If the value of the last write on x is v3 and a thread t executes  CAS x  v1 v2 r then the value of the last write on x
either remains the same or becomes v2.*)
lemma cas_possible_lv_lc:
  assumes "mem_structured ts"
 and "vbounded ts"
and "  \<lceil>x:v3\<rceil> ts "
and " step t ( CAS x  v1 v2 r) ts ts' "
shows "  \<lceil>x:v2\<rceil> ts' \<or>  \<lceil>x:v3\<rceil> ts'  "
  using assms
  apply (simp add: step_def)
  apply clarify
apply (case_tac " ts' = cas_succ_trans t ind x v1 v2 r ts")
  apply (metis assms(4) cas_lc cas_suc_reg compval_def zero_neq_one)
  apply simp
 apply (subgoal_tac " memory ts = memory ts'")
   prefer 2
   apply simp
  apply (subgoal_tac " last_entry ts x = last_entry ts' x")
   prefer 2
   apply (simp(no_asm) add: last_entry_def)
   apply (simp(no_asm) add: last_entry_set_def)
   apply presburger
  by presburger


lemma cas_fail_opts_ni :
 assumes "mem_structured ts"
 and "vbounded ts"
and " ts' = cas_fail_trans t ind x v1 v2 r ts"
and " x \<noteq> y"
shows " OPTS ts'  y =  OPTS ts  y"
  using assms

  apply (subgoal_tac " maxvp ts y = maxvp  ts' y ")
  prefer 2
  apply (simp add:cas_fail_trans_def Let_def)

   apply (subgoal_tac " memory ts = memory ts' ")
  prefer 2
  apply simp
  by (simp add: OPTS_def notoverwritten_def)



lemma ca_succ_opts_daddr_ni: 
  assumes "mem_structured ts"
  and "vbounded ts"
 and "x \<noteq> y"
  and  " ts' = cas_succ_trans t ind x v1 v2 r ts"
shows " OPTS ts'  y  =  OPTS ts y "
using assms 
  apply (simp add: step_def)
  apply (simp add: OPTS_def)
  apply (simp add: notoverwritten_def)

  apply (subgoal_tac " maxvp ts' y = maxvp ts y")
  prefer 2
  apply (simp add:  cas_succ_trans_def)

apply (subgoal_tac " (\<forall>i.(i\<ge>0\<and>i<length(memory ts)) \<longrightarrow> ( memory ts)!i =( memory ts')!i) ")
    prefer 2
  using mem_l_casucc apply blast
   apply (subgoal_tac" length( memory ts) \<notin>  OPTS ts'  y")
  prefer 2
   apply (simp add: OPTS_def)
  apply clarify
  apply safe
  apply (simp_all)
                    apply (metis Msg.sel(1) less_antisym nth_append_length)
              apply (metis Msg.sel(1) less_Suc_eq nth_append_length)
             apply (metis Msg.sel(1) less_Suc_eq nth_append_length)
            apply (metis Msg.discI not_less_less_Suc_eq nth_append_length)
  using less_Suc_eq apply auto[1]
          apply (metis Msg.distinct(1) less_antisym nth_append_length)
         apply (metis Msg.discI less_Suc_eq nth_append_length)
  apply (metis Msg.sel(1) not_less_less_Suc_eq nth_append_length)
   apply (metis Msg.sel(1) not_less_less_Suc_eq nth_append_length) 
   apply (metis Msg.sel(1) assms(3) less_SucE nth_append_length)
     apply (metis Msg.sel(1) less_Suc_eq nth_append_length)
    apply (metis Msg.sel(1) less_Suc_eq nth_append_length)
   apply (metis Msg.sel(1) less_Suc_eq nth_append_length)
  by (metis Msg.sel(1) less_Suc_eq nth_append_length)


lemma cas_opts_ni:
 assumes "mem_structured ts"
 and "vbounded ts"
and " step t ( CAS x  v1 v2 r) ts ts' "
and " x \<noteq> y"
shows " OPTS ts'  y =  OPTS ts  y"
  using assms
  apply (simp add: step_def)
  apply clarify
  apply (case_tac " ts' = cas_fail_trans t ind x v1 v2 r ts")
  apply (meson cas_fail_opts_ni)
  by (meson ca_succ_opts_daddr_ni)


(*rule [CS3]: An execution of CAS on address x does not affect the observable  persistent values of any other address.*)
lemma  cas_opv_ni:
 assumes "mem_structured ts"
 and "vbounded ts"
and " step t ( CAS x  v1 v2 r) ts ts' "
and " x \<noteq> y"
shows "[y]\<^sub>P ts = [y]\<^sub>P ts'"
  using assms
  apply (subgoal_tac " OPTS ts'  y =  OPTS ts  y")
  prefer 2
   apply (simp add: cas_opts_ni)
  apply (simp add: step_def)
  apply clarify
  apply (case_tac " ts' = cas_fail_trans t ind x v1 v2 r ts")
   apply simp
 apply (subgoal_tac " mapval (OPTS ts  y)  (memory ts') = mapval (OPTS ts' y) (memory ts )")
 prefer 2
    apply (subgoal_tac "\<forall> i. i \<in> (OPTS ts  y) \<longrightarrow> i < length (memory ts)")
    prefer 2
    apply (simp add: OPTS_def)
  apply (unfold mapval_def )
  using mem_l_casucc apply fastforce
  by presburger
 


lemma cas_succ_last_entry_in_ots :
  assumes "mem_structured ts"
and "vbounded ts"
and " step t ( CAS lx  v1 v2 r) ts ts' "
and " regs ts' t r = 1"
 and "\<forall> ti l. comploc ( (memory ts)!(coh ts ti l)) l = l"
shows "  last_entry ts' x \<in> OTS ts' t x "
  using assms
 apply (subgoal_tac"vbounded ts' ")
     prefer 2
  using vbounded_preserved apply blast
 apply (subgoal_tac" mem_structured ts' ")
   prefer 2
  using mem_structured_preserved apply blast
  apply (simp add: step_def)
  apply clarify
  apply (subgoal_tac  "  ts' = cas_succ_trans t ind lx v1 v2 r ts ")
   prefer 2
   apply (metis cas_fail_reg n_not_Suc_n)
apply (unfold  OTS_def)
    apply (unfold  OTSF_def)
  apply clarify
 apply (rule_tac x=" last_entry ts' x" in exI)
   apply (intro conjI impI)
  prefer 6
  using last_entry_notoverwritten apply blast
       apply blast
      apply linarith
  using last_entry_bounded apply blast
    apply (metis last_entry_loc)
 apply (subgoal_tac"notoverwritten ts' (vrnew ts' t) (last_entry ts' x) x")
     prefer 2
  using last_entry_notoverwritten apply blast

apply (subgoal_tac " coh ts' t x  < length(memory ts') ")
   prefer 2
          apply (subgoal_tac "vbounded ts'")
      prefer 2
    apply blast
          apply (unfold vbounded_def)
          apply meson
         apply (subgoal_tac " 0 \<le>  coh ts' t x")
          prefer 2
   apply blast
 apply (subgoal_tac " \<forall> i. i \<in> last_entry_set ts'  x \<longrightarrow> i \<le> last_entry ts' x")
   prefer 2
   apply (simp add: last_entry_def )
   apply (metis Max.in_idem finite_last_entry_set max_def order_refl)
  apply (subgoal_tac " \<forall> i. 0\<le> i \<and> i < length (memory ts') \<and>  comploc ( memory ts'!i) x = x\<longrightarrow>  i \<in> last_entry_set ts'  x ")
   prefer 2
   apply (simp add: last_entry_set_def)
  using assms(2) assms(3) assms(5) coh_loc_rel_preserved by presburger


lemma cas_fail_last_entry_in_ots :
  assumes "mem_structured ts"
and "vbounded ts"
and " step t ( CAS lx  v1 v2 r) ts ts' "
and " regs ts' t r = 0"
 and "\<forall> ti l. comploc ( (memory ts)!(coh ts ti l)) l = l"
shows "  last_entry ts' x \<in> OTS ts' t x "
  using assms
 apply (subgoal_tac"vbounded ts' ")
     prefer 2
  using vbounded_preserved apply blast
 apply (subgoal_tac" mem_structured ts' ")
   prefer 2
  using mem_structured_preserved apply blast
  apply (simp add: step_def)
  apply clarify
  apply (subgoal_tac  "  ts' = cas_fail_trans t ind lx v1 v2 r ts ")
   prefer 2
   apply (metis cas_suc_reg zero_neq_one)
apply (unfold  OTS_def)
    apply (unfold  OTSF_def)
  apply clarify
 apply (rule_tac x=" last_entry ts' x" in exI)
   apply (intro conjI impI)
  prefer 6
  using last_entry_notoverwritten apply blast
       apply blast
      apply linarith
  using last_entry_bounded apply blast
    apply (metis last_entry_loc)
 apply (subgoal_tac"notoverwritten ts' (vrnew ts' t) (last_entry ts' x) x")
     prefer 2
  using last_entry_notoverwritten apply blast

apply (subgoal_tac " coh ts' t x  < length(memory ts') ")
   prefer 2
          apply (subgoal_tac "vbounded ts'")
      prefer 2
    apply blast
          apply (unfold vbounded_def)
          apply meson
         apply (subgoal_tac " 0 \<le>  coh ts' t x")
          prefer 2
   apply blast
 apply (subgoal_tac " \<forall> i. i \<in> last_entry_set ts'  x \<longrightarrow> i \<le> last_entry ts' x")
   prefer 2
   apply (simp add: last_entry_def )
   apply (metis Max.in_idem finite_last_entry_set max_def order_refl)
  apply (subgoal_tac " \<forall> i. 0\<le> i \<and> i < length (memory ts') \<and>  comploc ( memory ts'!i) x = x\<longrightarrow>  i \<in> last_entry_set ts'  x ")
   prefer 2
   apply (simp add: last_entry_set_def)
  using assms(2) assms(3) assms(5) coh_loc_rel_preserved by presburger


lemma last_entry_succ_daddr_preserved:
 assumes "mem_structured ts"
 and "vbounded ts"
and " x \<noteq> y"
and "  ts' = cas_succ_trans t ind x v1 v2 r ts"
shows "  last_entry ts'  y = last_entry ts  y"
  using assms
  apply (subgoal_tac " (\<forall>i.(i\<ge>0\<and>i<length(memory ts)) \<longrightarrow> ( memory ts)!i =( memory ts')!i) ")
   prefer 2
  using mem_l_casucc apply blast
  apply (simp add: last_entry_def)
  apply (subgoal_tac " last_entry_set ts' y = last_entry_set ts y")
   prefer 2
   apply (subgoal_tac " length(memory ts) \<notin> last_entry_set ts' y")
    prefer 2
    apply (simp add: last_entry_set_def)
   apply (simp add: last_entry_set_def)
  using less_Suc_eq apply force
  by simp



lemma cas_lv_ots_daddr_lc:
  assumes "mem_structured ts"
 and "vbounded ts"
and " x \<noteq> y"
and " OTS ts t' y  = {w} "
and " step t ( CAS x  v1 v2 r) ts ts' "
 and "\<forall> ti l. comploc ( (memory ts)!(coh ts ti l)) l = l"
 and "  \<lceil>y\<rceil>\<^sub>t' ts"
shows "    (  OTS ts' t y  = {w}  \<and>  regs ts' t r = 1 ) \<or>  regs ts' t r= 0   "
  using assms
  apply (simp add: step_def)
  apply clarify
  apply (subgoal_tac " vbounded ts' ")
   prefer 2
  using assms(5) vbounded_preserved apply blast
 apply (subgoal_tac " maxcoh ts t < length (memory ts)")
   prefer 2
   apply (simp add: vbounded_def)
 apply (subgoal_tac " vrnew ts t < length (memory ts)")
      prefer 2
   apply (simp add: vbounded_def)
  apply (subgoal_tac "  regs ts' t r = 1  \<or>   regs ts' t r= 0 ")
   prefer 2
   apply (metis assms(5) cas_lc)
 apply (case_tac "  regs ts' t r = 1 ")
    apply (subgoal_tac"  ts' = cas_succ_trans t ind x v1 v2 r ts")
    prefer 2
    apply (metis cas_fail_reg zero_neq_one)
 apply (subgoal_tac "  last_entry ts'  y \<in>   OTS ts' t y ")
    prefer 2
  using assms(5) assms(6) cas_succ_last_entry_in_ots apply blast
 apply (subgoal_tac "  last_entry ts'  y =   last_entry ts  y ")
    prefer 2
    apply (metis last_entry_succ_daddr_preserved)
(* apply (subgoal_tac " OTS ts' t y  = { last_entry ts  y} ")*)
  apply (subgoal_tac " vrnew ts' t  = length (memory ts)")
     prefer 2
    apply (simp add:  cas_succ_trans_def)
(* apply (subgoal_tac " \<nexists> n.( n \<noteq> last_entry ts  y \<and>0 \<le> n \<and> n < length(memory ts') \<and> n \<in>  OTS ts' t y)")  *)
 apply (subgoal_tac " \<forall> n.( n \<noteq> last_entry ts'  y \<and>0 \<le> n \<and> n < length(memory ts') \<longrightarrow> n \<notin>  OTS ts' t y)") 
   prefer 2
    apply clarify
    apply (simp add: last_entry_def)
    apply (simp add: OTS_def)
    apply (simp add: OTSF_def)
  apply (case_tac "n > last_entry ts'  y")
  (*  apply (subgoal_tac " comploc ( (memory ts')! n) y \<noteq> y")
      prefer 2*)
apply (subgoal_tac " \<forall> j. ( 0 \<le> j \<and> j < length (memory ts')  \<and> comploc (memory ts' ! j) y = y) \<longrightarrow> j \<in>  last_entry_set  ts'  y")
      prefer 2
       apply (simp add: last_entry_set_def)
    apply (subgoal_tac " \<forall> i. i \<in>  last_entry_set  ts'  y \<longrightarrow> i \<le>  last_entry ts'  y")
      prefer 2
      apply (subgoal_tac " last_entry_set ts' y \<noteq> {}")
       prefer 2
       apply (metis cas_succ_wfs last_entry_set_not_empty)
 apply (subgoal_tac " finite ( last_entry_set ts' y )")
       prefer 2
        apply (metis finite_last_entry_set)
       apply (metis Max_ge last_entry_def)
      apply (subgoal_tac "(n >  last_entry ts'  y \<and>  0 \<le> n \<and> n < length (memory ts')) \<longrightarrow> ( comploc (memory ts' ! n) y \<noteq> y)")
       prefer 2
       apply (metis le_antisym nat_less_le)
     apply (subgoal_tac "(n >  last_entry ts'  y \<and>  0 \<le> n \<and> n < length (memory ts'))  \<longrightarrow> n \<notin>  OTS ts' t y   ")
      prefer 2
      apply (simp add: OTS_def)
     apply (simp add: OTSF_def)
    apply (subgoal_tac " n \<noteq>  last_entry ts'  y ")
     prefer 2
  apply (metis last_entry_def)
    apply (subgoal_tac "(n <  last_entry ts'  y \<and>  0 \<le> n \<and> n < length (memory ts')) \<longrightarrow> ( \<not> notoverwritten ts'  (length (memory ts)) n y)")
     prefer 2
     apply (subgoal_tac " \<exists> j.( n < j \<and> j \<le> length (memory ts)) \<and>  ( comploc (memory ts' ! j) y = y)")
  prefer 2
      apply clarify
      apply (rule_tac x=" last_entry ts' y " in exI)
      apply (intro conjI)
        apply (meson nat_neq_iff)
       apply (metis last_entry_def not_le_imp_less not_less_eq)
      apply (meson cas_succ_wfs last_entry_loc)

     apply (simp add: notoverwritten_def)
  apply (metis (no_types, lifting) last_entry_bounded last_entry_def last_entry_succ_daddr_preserved less_Suc_eq_0_disj less_Suc_eq_le linorder_neqE_nat mem_l mem_structured_def store_add)
    apply (simp add: OTS_def)
   apply (subgoal_tac "  OTS ts' t y = {last_entry ts' y}")
    prefer 2
    apply (subgoal_tac " \<forall>i. i \<in> OTS ts' t y \<longrightarrow>   0 \<le> i \<and> i < length (memory ts')")
  prefer 2
    apply (simp add: OTS_def)
     apply (simp add: OTSF_def)
    apply blast
   apply (metis One_nat_def)
  by (metis less_numeral_extra(3))


(*rule [CP4]: Given x different to y, if thread's t view of y is the
 last write on y  and  the value of this write is v  then after executing the CAS  either the CAS succeeds, so r=1  and t's  view of y becomes v or the CAS fails and r=0.*)
lemma cas_lv_daddr_lc:
  assumes "mem_structured ts"
 and "vbounded ts"
and " x \<noteq> y"
and " \<lceil>y, v\<rceil>\<^sub>t'  ts "
and " step t ( CAS x  v1 v2 r) ts ts' "
 and "\<forall> ti l. comploc ( (memory ts)!(coh ts ti l)) l = l"
shows "    ( [y]\<^sub>t  ts' = {v}  \<and>  regs ts' t r = 1 ) \<or>  regs ts' t r= 0   "   
  using assms
  apply (subgoal_tac "  ( ( OTS ts' t y  =  OTS ts t' y)  \<and>  regs ts' t r = 1 ) \<or>  regs ts' t r= 0 ")
   prefer 2
   apply (metis cas_lv_ots_daddr_lc)
  apply (simp add: step_def)
  apply clarify
 apply (subgoal_tac "  regs ts' t r = 1  \<or>   regs ts' t r= 0 ")
   prefer 2
   apply (metis assms(5) cas_lc)
 apply (case_tac "  regs ts' t r = 1 ")
    apply (subgoal_tac"  ts' = cas_succ_trans t ind x v1 v2 r ts")
    prefer 2
    apply (metis cas_fail_reg zero_neq_one)
 apply (subgoal_tac " mapval (OTS ts t'  y)  (memory ts') = mapval (OTS ts t'  y) (memory ts )")
 prefer 2
    apply (subgoal_tac "\<forall> i. i \<in> (OTS ts t'  y) \<longrightarrow> i < length (memory ts)")
    prefer 2
     apply simp
    apply (unfold mapval_def )
  using last_entry_bounded apply blast
    apply (subgoal_tac " mapval (OTS ts t'  y) (memory ts ) = mapval (OTS ts' t  y) (memory ts )")
     prefer 2
     apply (metis zero_neq_one)
  apply (smt (z3) assms(5) cas_nov_daddr_ni cas_succ_ots_daddr_dt image_insert image_is_empty mapval_def mem_Collect_eq neq0_conv singleton_conv2)
  apply (metis zero_less_iff_neq_zero)
  by simp

lemma cas_fail_ots__sub:
  assumes "mem_structured ts"
  and "vbounded ts"
and "  ts' = cas_fail_trans t ind x v1 v2 r ts "
and "y \<noteq> x"
shows  "OTS ts' t' y   \<subseteq> OTS ts t'  y"
  using assms
apply (subgoal_tac " memory ts = memory ts' ")
   prefer 2
   apply auto[1]
 apply (subgoal_tac " vrnew ts t'  \<ge> vrnew ts t'  ")
    prefer 2
   apply (simp add: cas_fail_trans_def)
  apply (subgoal_tac " coh ts t' y = coh ts t' y ")
    prefer 2
   apply (simp add: cas_fail_trans_def)
   apply (simp add: OTS_def )
   apply (simp add: OTSF_def )
   apply (simp add: notoverwritten_def)
  apply safe
         apply (simp add: cas_fail_trans_def split:if_splits)
        apply (simp add: cas_fail_trans_def split:if_splits)
       apply (simp add: cas_fail_trans_def split:if_splits)
      apply (simp add: cas_fail_trans_def split:if_splits)
  using le_max_iff_disj apply auto[1]
       apply blast
      apply blast
     apply (simp add: cas_fail_trans_def split:if_splits)
    apply (simp add: cas_fail_trans_def split:if_splits)
   apply (simp add: cas_fail_trans_def split:if_splits)
  by (simp add: cas_fail_trans_def split:if_splits)


(*rule [CS4]: The value of the last write on y remains the same after executing a CAS on a different address.*)
lemma cas_ots_lm_ni:
  assumes "mem_structured ts"
  and "vbounded ts"
and " step t ( CAS x  v1 v2 r) ts ts' "
  and "x \<noteq> y"
  and "  \<lceil>y\<rceil>\<^sub>t'  ts "
and "\<forall> ti l.  OTS  ts ti  l \<noteq> {}"
and  "\<forall>ti l. comploc ( (memory ts)!(coh ts ti l)) l = l"
shows "  \<lceil>y\<rceil>\<^sub>t'  ts'  "
  using assms
  apply (simp add: step_def)
  apply clarify
  apply(case_tac " ts' =  cas_succ_trans t ind x v1
            v2 r ts")
  apply (subgoal_tac " last_entry ts y = last_entry ts' y") 
   prefer 2
  using last_entry_succ_daddr_preserved apply force
   apply (subgoal_tac" OTS ts' t' y \<subseteq>  OTS ts t' y ")
    prefer 2
    apply (metis cas_succ_ots_daddr_dt cas_succ_ots_daddr_st dual_order.eq_iff)
   apply (metis OTS_notempty_preserved assms(3) assms(7) subset_singletonD)
  apply (subgoal_tac " memory ts = memory ts'")
  prefer 2
  apply simp
 apply (subgoal_tac " last_entry ts y = last_entry ts' y") 
   prefer 2
   apply  (simp add: last_entry_def)
   apply  (simp add: last_entry_set_def)
  apply (subgoal_tac" OTS ts' t' y \<subseteq>  OTS ts t' y ")
   prefer 2
   apply (metis cas_fail_ots__sub)
  by (metis OTS_notempty_preserved assms(3) assms(7) subset_singletonD)
  






end