(*Title: View-Based Owicki-Gries Reasoning for Persistent x86-TSO 
  Authors: Eleni Bila, Brijesh Dongol, Ori Lahav, Azalea Raad, John Wickerson*)

theory Flushopt_Rules
imports   "../State" "../Language"  "../Assertions"  "../Wellformedness"
begin

(* The lemmas that begin with the  comment "rule [name]: __" are the proof rules regarding the flushopt instruction. The
name corresponds to the name of the rule in the paper. 
   All the other lemmas are auxiliary and are used for proving the proof rules.
   Rules that end with "lc" concern mostly local correctness.
   Rules that end with "ni" concern  mostly  non-interference.
   Rules that end with "gen" concern non-interference and local correctness.
   Rules that end with "rel" concern relations between view sets. *) 

lemma [simp]:
  assumes "mem_structured ts"
  and " i \<in>  OTS ts ti x"
shows " 0 \<le> i  \<and> i < length(memory ts) "
  using assms apply (simp add:  mem_structured_def OTS_def OTSF_def )
  done


lemma [simp]:
  assumes "mem_structured ts"
 and " 0 \<le> i  \<and> i < length(memory ts) "
 and "memory ts ! i = Init_Msg"
shows "i = 0"
  using assms(1)  apply ( unfold  mem_structured_def   )
  apply (subgoal_tac"  length (memory ts) \<noteq>  0 " )
   prefer 2
   apply blast
  using assms(2) assms(3) le_neq_implies_less by blast



lemma flo_oats_ots_rel:
  assumes "mem_structured ts"
  and "vbounded ts"
  and  "step ti ( Flush_opt  x ) ts ts'"
  and " OTS ts  ti  x\<noteq> {}"
 and  "\<forall>ti l. comploc ( (memory ts)!(coh ts ti l)) l = l"
shows "  OATS ts' ti x \<subseteq> OTS ts  ti  x "
using assms 
  apply (simp add: step_def )
  apply (subgoal_tac "\<forall>k.  k \<in> OATS ts' ti x \<longrightarrow> k \<in> OTS ts  ti  x" )
  prefer 2
apply (simp add: OATS_def OTS_def)
  apply (simp add:  OTSF_def notoverwritten_def)
  apply clarify
  apply(intro allI impI conjI, elim conjE)
 apply (subgoal_tac " k = 0")
     prefer 2
     apply auto[1]
    apply simp
 apply (subgoal_tac " coh ts ti x \<le>  vpasync (flush_opt_trans ti x ts) ti x ")
     prefer 2
     apply (simp add: flush_opt_trans_def)
    apply (meson dual_order.strict_trans2 mem_structured_def neq0_conv)
   apply simp
   apply (subgoal_tac " vrnew ts ti \<le>  vpasync ts' ti x")
  prefer 2
   apply (subgoal_tac "  vpasync ts' ti x \<ge> vpready ts ti")
    prefer 2
    apply (simp add: flush_opt_trans_def )
   apply (subgoal_tac " vrnew ts ti \<le> vpready ts ti ")
    prefer 2
  using assms(2) vbounded_def apply blast
      apply linarith
     apply simp
    apply simp
 apply (subgoal_tac " coh ts ti x \<le>  vpasync (flush_opt_trans ti x ts) ti x ")
     prefer 2
     apply (simp add: flush_opt_trans_def)
 apply (subgoal_tac "comploc (( memory ts)! (coh ts ti x)) x = x")
     prefer 2
  using assms(5) apply blast
     apply( elim conjE)
    apply simp
     apply (simp add: mem_structured_def )
    apply (meson le0 leI le_less_trans)
  apply (subgoal_tac " vrnew ts ti \<le>  vpasync ts' ti x")
  prefer 2
   apply (subgoal_tac "  vpasync ts' ti x \<ge> vpready ts ti")
    prefer 2
    apply (simp add: flush_opt_trans_def )
   apply (subgoal_tac " vrnew ts ti \<le> vpready ts ti ")
    prefer 2
  using assms(2) vbounded_def apply blast
      apply linarith
   apply simp
  by (simp add: subsetI) 


(*rule[OP]: After a thread t executes a flushopt on address x, the set of its observable asynchronous
values (OAV) of x for t, is a subset of the set of its observable values (OV) *)
lemma flo_oav_ov_rel:
  assumes "mem_structured ts"
  and "vbounded ts"
  and  "step t ( Flush_opt  x ) ts ts'"
  and " OTS ts  t  x\<noteq> {}"
  and  "\<forall>ti l. comploc ( (memory ts)!(coh ts ti l)) l = l"
shows "  [x]\<^sup>A\<^sub>t ts'  \<subseteq>  [x]\<^sub>t ts"
  using assms
 apply (simp add: step_def )
  apply (subgoal_tac " OATS ts' t  x \<subseteq> OTS ts  t  x" )
   prefer 2
  using assms(3) assms(5) flo_oats_ots_rel apply blast
  apply (simp add: mapval_def)
  by auto



lemma [simp]:
 assumes "mem_structured ts"
and "i = 0"
shows "(memory ts ) ! i =  Init_Msg"
  using assms
  apply (simp add: mem_structured_def vbounded_def )
  done 

lemma vrnew_flo:
 assumes "mem_structured ts"
  and "vbounded ts"
and " ts' = flush_opt_trans t x ts"
shows " vrnew ts' t' =  vrnew ts t'"
  using assms 
  by(simp add: flush_opt_trans_def)

lemma coh_flo:
 assumes "mem_structured ts"
  and "vbounded ts"
and " ts' = flush_opt_trans t x ts"
shows " coh ts' t' y =  coh ts t' y"
 using assms 
  by(simp add: flush_opt_trans_def)
 
lemma mem_flo:
 assumes "mem_structured ts"
  and "vbounded ts"
and " ts' = flush_opt_trans t x ts"
shows " memory ts =  memory ts'"
  using assms 
  by simp

(*rule[OS1]: After executing a flushopt, the observable  values (OV) for all threads/addresses
remain the same*)
lemma flo_ov_ni: 
  assumes "mem_structured ts"
  and "vbounded ts"
 and  "step t' ( Flush_opt x) ts ts'"
shows "   [y]\<^sub>t ts=  [y]\<^sub>t ts'"
using assms 
  apply (simp add: step_def)
  apply (subgoal_tac " OTS ts t y = OTS ts' t y")
  prefer 2
   apply (subgoal_tac " vrnew ts' t =  vrnew ts t")
    prefer 2
  apply (simp add: vrnew_flo)
  apply (subgoal_tac " coh  ts' t y =  coh  ts t y")
    prefer 2
  using coh_flo apply blast
    apply (subgoal_tac " memory ts = memory ts'")
    prefer 2
  apply simp
   apply (simp add: OTS_def)
   apply (simp add: OTSF_def)
   apply (smt Collect_cong mem_flo notoverwritten_def)
  apply (simp add: mapval_def)
  done


lemma vpcommit_flo:
 assumes "mem_structured ts"
  and "vbounded ts"
and " ts' = flush_opt_trans t x ts"
shows " vpcommit ts t =  vpcommit ts' t"
  using assms 
  apply (simp add:  flush_opt_trans_def)
  done



lemma vp_commit_flo_gen:
assumes  "mem_structured ts"                                  
and   " ts' = flush_opt_trans t x ts" 
shows " ( \<forall> ti. ( vpcommit ts' ti l =  vpcommit ts ti l)  )"
 using assms
  apply (simp add: flush_opt_trans_def)
  done  


lemma vpasync_flo_daddr:
 assumes "mem_structured ts"
  and "vbounded ts"
and " ts' = flush_opt_trans t x ts"
and " x \<noteq> y"
shows " ( vpasync ts t' y =   vpasync ts' t' y )"
 using assms 
  apply (simp add:  flush_opt_trans_def)
  done


lemma flo_mem: 
assumes  "mem_structured ts"                                  
and " ts' = flush_opt_trans t x ts" 
shows "memory ts = memory ts' "
 using assms
  apply (simp add: flush_opt_trans_def)
  done

lemma maxvp__flo_gen:
 assumes "mem_structured ts"
  and "vbounded ts"
and " ts' = flush_opt_trans t x ts"
shows " (\<forall> t.  maxvp ts t =  maxvp ts' t)"
 using assms 
  apply (simp add:  flush_opt_trans_def)
  done



lemma flo_opts_ni:
  assumes "mem_structured ts"
  and "vbounded ts"
 and  "step t ( Flush_opt x) ts ts'"
shows "    OPTS ts y  = OPTS ts' y  "
using assms 
  apply (simp add: step_def)
  apply (subgoal_tac "memory ts = memory ts' ")
  prefer 2
  apply auto[1]
  apply (subgoal_tac " \<forall> ti.( maxvp ts ti =  maxvp ts' ti) ")
  prefer 2
  using maxvp__flo_gen apply blast
  apply (simp add: OPTS_def)
  by (metis (no_types, lifting) flo_mem notoverwritten_def)

(*rule[OS2]: After executing a flushopt, the observable persistent  values (OPV) for all addresses
remain the same*)
lemma flo_opv_ni:
  assumes "mem_structured ts"
  and "vbounded ts"
 and  "step t ( Flush_opt x) ts ts'"
shows "   [y]\<^sub>P ts'  =  [y]\<^sub>P ts "
  using assms
  apply (subgoal_tac " OPTS ts y  = OPTS ts' y  ")
   prefer 2
   apply (simp add: flo_opts_ni)
  apply (simp add: mapval_def)
  apply (subgoal_tac " memory ts   = memory ts'  ")
   apply simp
  apply (simp add: step_def)
  done

(*rule: flo_oav_ov_rel rule for singleton OV*)
corollary flo_oav_ov_rel_s:
  assumes "mem_structured ts"
  and "vbounded ts"
  and  "step t ( Flush_opt  x ) ts ts'"
  and " OTS ts  t  x\<noteq> {}"
  and  "\<forall>ti l. comploc ( (memory ts)!(coh ts ti l)) l = l"
 and "  [x]\<^sub>t ts = {w} "
 and " (\<forall>  addr t. [addr]\<^sup>A\<^sub>t ts \<noteq> {})"
shows "  [x]\<^sup>A\<^sub>t ts' = {w}"
  using assms
 apply (subgoal_tac "  [x]\<^sup>A\<^sub>t ts' \<noteq> {}")
   prefer 2
  using OAV_notempty_preserved  apply blast
  apply (simp add: step_def )
  by (metis assms(3) assms(5) flo_mem flo_oav_ov_rel subset_singletonD)

lemma flo_oats_sub:
  assumes "mem_structured ts"
  and "vbounded ts"
  and  "step t ( Flush_opt  x ) ts ts'"
shows  "OATS ts' t    x \<subseteq> OATS ts t  x"
  using assms
  apply (simp add: step_def)

  apply (subgoal_tac " memory ts = memory ts' ")
   prefer 2
   apply simp
  apply (subgoal_tac " vpasync ts' t  x \<ge> vpasync ts t x ")
   prefer 2
   apply (simp add: flush_opt_trans_def Let_def)
  apply (simp add: OATS_def )
  by (smt (verit, ccfv_SIG) Collect_mono_iff dual_order.trans flo_mem notoverwritten_def)

(*rule [OP]: After a thread t executes a flushopt on x the observable asynchronous values for x might become less.*)
lemma fl_oav_sub:
  assumes "mem_structured ts"
  and "vbounded ts"
  and  "step t ( Flush_opt  x ) ts ts'"
shows  "  [x]\<^sup>A\<^sub>t ts'   \<subseteq>  [x]\<^sup>A\<^sub>t ts"
  using assms
  apply (subgoal_tac " OATS ts' t   x \<subseteq> OATS ts  t  x ")
   prefer 2
   apply (simp add: flo_oats_sub)
 apply (subgoal_tac " memory ts =memory ts' ")
   prefer 2
  apply (simp add: step_def)
  apply (simp add: mapval_def)
  by blast

(*rule [OS3]: The number of writes on address x with value v remain the same after executing a flushopt.*)
lemma  flo_oc_ni : 
  assumes "mem_structured ts"
  and "step t ( Flush_opt y) ts ts'"
  and "vbounded ts"
shows  " \<lparr>x,v\<rparr>  ts' =  \<lparr>x,v\<rparr>  ts "
  using assms
  apply (simp add: step_def)
  by (simp add:  oc_set_def)

lemma flo_oats_daddr_ni:
  assumes "mem_structured ts"
  and "vbounded ts"
 and " x \<noteq> y"
  and  "step t ( Flush_opt  x ) ts ts'"
shows  "OATS ts' t'    y = OATS ts t'  y"
  using assms
  apply (simp add: step_def)
 apply (subgoal_tac " memory ts = memory ts' ")
   prefer 2
   apply simp
  apply (subgoal_tac "  vpasync ts t' y =   vpasync ts' t' y")
   prefer 2
  using vpasync_flo_daddr apply blast
  apply (simp add: OATS_def)
  by (metis (no_types, lifting)  flo_mem notoverwritten_def)


(*rule:The set of the asynchronous values for an address y remain the same after a thread executes 
a flushopt on a different address.*)
lemma flo_oav_daddr_ni:
  assumes "mem_structured ts"
  and "vbounded ts"
 and " x \<noteq> y"
  and  "step t' ( Flush_opt  x ) ts ts'"
shows  " [y]\<^sup>A\<^sub>t ts' = [y]\<^sup>A\<^sub>t ts"
  using assms
  apply (subgoal_tac  "OATS ts' t'    y = OATS ts t'  y")
   prefer 2
  apply (simp add: flo_oats_daddr_ni)
  apply (simp add: mapval_def)
 apply (subgoal_tac " memory ts =memory ts' ")
   prefer 2
  apply (simp add: step_def)

  using flo_oats_daddr_ni by presburger




end