(*Title: View-Based Owicki-Gries Reasoning for Persistent x86-TSO 
  Authors: Eleni Bila, Brijesh Dongol, Ori Lahav, Azalea Raad, John Wickerson*)

theory Load_Rules
imports   "../State" "../Language"  "../Assertions"  "../Wellformedness"
begin

(* The lemmas that begin with the  comment "rule [name]: __" are the proof rules regarding the load instruction. The
name corresponds to the name of the rule in the paper. 
   All the other lemmas are auxiliary and are used for proving the proof rules.
   Rules that end with "lc" concern mostly local correctness.
   Rules that end with "ni" concern  mostly  non-interference.
   Rules that end with "gen" concern non-interference and local correctness.
   Rules that end with "rel" concern relations between view sets. *) 


lemma mem_ld:
 assumes "mem_structured ts"
  and "vbounded ts"
and " ts' = load_trans t ind x  ts r"
shows " memory ts =  memory ts'"
  using assms 
  by simp


(*rule [LP1]: A thread can only read values at address x that belong to its set of observable values for x (OV)*)
lemma ld_ov_lc:
  assumes "mem_structured ts"
 and "step t ( Load x r) ts ts'"
shows "regs ts' t r \<in> [x]\<^sub>t ts "
  using assms apply (simp add:  step_def load_trans_def Let_def  mem_structured_def  mapval_def )
  apply clarify
  apply (subgoal_tac " ind \<in> OTS ts t x")
   prefer 2
   apply blast
  apply (subgoal_tac " compval((memory ts)!ind) \<in> [x]\<^sub>t ts")
   prefer 2
  using mapval_def apply auto[1]
  apply (subgoal_tac " compval((memory ts)!ind) =  regs ts' t r")
   prefer 2
   apply (simp add: load_trans_def)
  by auto




lemma ld_cobts_lc : 
  assumes "mem_structured ts"
  and "step t ( Load x r) ts ts'"
 and "vbounded ts"
 and "   regs ts' t r = v "
  and " x \<noteq> y"
 and  "  COBTS x y v t ts = S "
shows "  OTS ts' t y \<subseteq> S"
  using assms
  apply (simp add: step_def )
  apply clarify
  apply (subgoal_tac "compval (memory ts' ! ind) = v")
   prefer 2
   apply (simp add: load_trans_def)
  apply (subgoal_tac " OTSF ts t y (compvrnew ts t ind x)   = OTS ts' t y ")
   prefer 2
   apply (subgoal_tac " compvrnew ts t ind x = vrnew ts' t ")
    prefer 2
    apply (simp add: compvrnew_def load_trans_def)
  using max.commute apply blast

  apply (simp add: load_trans_def)
   apply (simp add: OTS_def  mem_structured_def OTSF_def notoverwritten_def )
  apply (simp add: COBTS_def cond_ts_def)
  by fastforce   

lemma ld_cobts_lc_aux : 
  assumes "mem_structured ts"
  and "step t ( Load x r) ts ts'"
 and "vbounded ts"
 and "   regs ts' t r = v "
 and  "  COBTS x x v t ts = S "
shows "  OTS ts' t x \<subseteq> S"
  using assms
  apply (simp add: step_def )
  apply clarify
  apply (subgoal_tac "compval (memory ts' ! ind) = v")
   prefer 2
   apply (simp add: load_trans_def)
  apply (subgoal_tac " OTSF ts t x (compvrnew ts t ind x)   = OTS ts' t x ")
   prefer 2
   apply (subgoal_tac " compvrnew ts t ind x = vrnew ts' t ")
    prefer 2
    apply (simp add: compvrnew_def load_trans_def)
  using max.commute apply blast
    apply (simp add: load_trans_def)
   apply (simp add: OTS_def  mem_structured_def OTSF_def notoverwritten_def )
   apply safe
  apply (metis bot_nat_0.not_eq_extremum linorder_not_le max.cobounded2)
         apply (metis (no_types, lifting) bot_nat_0.not_eq_extremum eq_imp_le max.coboundedI2)
        apply (metis leI max.cobounded1 max.commute)
  apply (metis bot_nat_0.not_eq_extremum eq_imp_le less_nat_zero_code max.coboundedI2 not_le_imp_less)
  using order.trans apply linarith
     apply (meson le_trans)
  using order.trans apply linarith
  using le_trans apply linarith
 apply (simp add: COBTS_def cond_ts_def)
  by fastforce  

(*rule [LP2]: Transfer rule for conditional observation*)
lemma ld_cobv_lc : 
  assumes "mem_structured ts"
 and "step t ( Load x r) ts ts'"
 and " \<langle>x  v\<rangle>[y]\<^sub>t  ts = S"
and "vbounded ts"
and  " regs ts' t r = v  "
shows "  [y]\<^sub>t  ts' \<subseteq>  S"
 using assms
  apply (simp add: step_def )
  apply (subgoal_tac "  OTS ts' t y \<subseteq>   COBTS x y v t ts ")
  prefer 2
  using ld_cobts_lc 
  apply (metis assms(2) ld_cobts_lc_aux) 
  by (smt (z3) imageE image_eqI mapval_def mem_ld subset_eq)



lemma mem_ld_trans :
  assumes " ts' = load_trans t' ind y ts r"
  shows " memory ts = memory ts'"
  using assms 
  by (simp add: load_trans_def)
 

lemma coh_ld :
  assumes " ts' = load_trans t' ind y ts r"
and "t \<noteq> t'"
  shows " coh ts t y = coh ts' t y"
  using assms 
  apply (simp add: load_trans_def)
  done

lemma coh_ld_dif :
  assumes " ts' = load_trans t' ind y ts r"
and "t \<noteq> t'"
and "x \<noteq> y'"
  shows " coh ts t x = coh ts' t x"
  using assms 
  apply (simp add: load_trans_def)
  done
 


(*rule [LS1]: After a thread t executes a load of address x, the observable values (OV) for x
remain the same for threads different from t.*)
lemma ld_ov_ni:
  assumes "mem_structured ts"
  and "vbounded ts"
 and "t \<noteq> t'"
  and "step t' ( Load x r) ts ts'"
shows "    [y]\<^sub>t ts  =   [y]\<^sub>t ts'"
  using assms 
  apply (simp add: step_def)
  apply (subgoal_tac " OTS ts t y = OTS ts' t y")
  prefer 2
  apply (case_tac " x= y")
    apply (simp add: OTS_def  )
    apply (subgoal_tac " vrnew ts t = vrnew ts' t")
     prefer 2
     apply (simp add: load_trans_def)
     apply auto[1]
apply (simp add: OTSF_def)
    apply (subgoal_tac "memory ts = memory ts'")
    prefer 2
  using mem_ld apply blast
  apply (subgoal_tac " coh ts t y =  coh ts' t y")
  prefer 2
     apply (metis coh_ld_dif)

  using notoverwritten_def apply presburger
  
   apply (simp add: OTS_def)
 apply (subgoal_tac " vrnew ts t = vrnew ts' t")
     prefer 2
     apply (simp add: load_trans_def)
    apply auto[1]
apply (simp add: OTSF_def)
    apply (subgoal_tac "memory ts = memory ts'")
    prefer 2
  using mem_ld apply blast
  apply (subgoal_tac " coh ts t y =  coh ts' t y")
  prefer 2
    apply (metis coh_ld_dif)
  apply (smt Collect_cong notoverwritten_def)
  apply (simp add: mapval_def)
  by auto

lemma coh_ld_st_sadrr :
   assumes "mem_structured ts"
  and "vbounded ts"
  and "step t ( Load y r) ts ts'"
  shows " coh ts t y  \<le> coh ts' t y"
  using assms 
   apply (simp add: step_def)
  apply (simp add: load_trans_def Let_def)
  apply clarify
  apply (case_tac "ind = coh ts t y")
   apply simp
  apply (simp add: OTS_def)
  apply (simp add: OTSF_def)
  done

lemma coh_ld_st_dadrr: 
   assumes "mem_structured ts"
  and "vbounded ts"
  and "step t ( Load x r) ts ts'"
   and " x \<noteq> y"
 shows " coh ts t y  = coh ts' t y"
  using assms 
   apply (simp add: step_def)
  apply (simp add: load_trans_def Let_def)
    apply clarify
  apply (case_tac "ind = coh ts t y")
 apply simp
 apply (simp add: OTS_def)
  done


lemma vrnew_ld_st_sadrr :
   assumes "mem_structured ts"
  and "vbounded ts"
  and "step t ( Load y r) ts ts'"
shows " vrnew ts t   \<le> vrnew  ts' t "
 using assms 
   apply (simp add: step_def)
  apply (simp add: load_trans_def Let_def)
   apply clarify
  apply (case_tac "ind = coh ts t y")
  apply simp
  apply (simp add: OTS_def)
  done

lemma vrnew_ld_st_dadrr :
   assumes "mem_structured ts"
  and "vbounded ts"
  and "step t ( Load x r) ts ts'"
   and " x \<noteq> y"
 shows " vrnew ts t   \<le> vrnew  ts' t "
  using assms(1) assms(2) assms(3) vrnew_ld_st_sadrr by auto


(*rule: If a  value k is not included to the observable values (OV) for 
address y  for a thread, it will also not be included after executing a load *)
lemma ld_nov_ni:
  assumes "mem_structured ts"
  and "vbounded ts"
  and "step t' ( Load x r) ts ts'"
  and "  k\<notin>  [y]\<^sub>t ts  "
shows "  k\<notin>  [y]\<^sub>t ts'"
 using assms 
  apply (simp add: step_def)
  apply (case_tac "t = t'")
  prefer 2
  using assms(3) ld_ov_ni apply blast
  apply (subgoal_tac" OTS ts' t' y \<subseteq> OTS ts t' y")
   prefer 2
  apply (case_tac" x = y")
   apply (subgoal_tac " coh ts t' y \<le> coh ts' t' y")
     prefer 2
  using assms(3) coh_ld_st_sadrr apply auto[1]
    apply (subgoal_tac "  vrnew ts t'   \<le> vrnew  ts' t' ")
     prefer 2
  using assms(3) vrnew_ld_st_sadrr apply blast
    apply (simp add: OTS_def )
    apply (simp add: OTSF_def notoverwritten_def )
    apply (subgoal_tac "  memory ts   =  memory  ts' ")
     prefer 2
  using mem_ld apply blast            
    apply clarify
  apply (intro conjI impI)
  apply presburger
  apply linarith
  apply (metis order_trans)
  apply presburger
  apply presburger
  apply linarith
    apply (metis order_trans) 
  apply (subgoal_tac " coh ts t' y = coh ts' t' y")
    prefer 2
  using assms(3) coh_ld_st_dadrr apply auto[1]
 apply (subgoal_tac "  vrnew ts t'   \<le> vrnew  ts' t' ")
    prefer 2
  using assms(3) vrnew_ld_st_sadrr apply auto[1]
  apply (simp add: OTS_def )
    apply (simp add: OTSF_def notoverwritten_def )
    apply (subgoal_tac "  memory ts   =  memory  ts' ")
     prefer 2
  using mem_ld apply blast 
  apply clarify
   apply (intro conjI impI)
         apply presburger
        apply blast
       apply (metis order_trans)
      apply presburger
     apply presburger
    apply linarith
   apply (metis order_trans)
  apply (subgoal_tac " \<forall>i.  i \<notin> OTS ts t' y \<longrightarrow>  i \<notin> OTS ts' t' y ")
  prefer 2
   apply auto[1]
  apply (simp add: mapval_def)
  by auto


corollary   ld_cobts_lc_s : 
  assumes "mem_structured ts"
  and "step t ( Load x r) ts ts'"
 and "vbounded ts"
 and "   regs ts' t r = v "
  and " x \<noteq> y"
 and  "  COBTS x y v t ts = {w} "
and "  (\<forall> ti addr. OTS   ts ti  addr \<noteq> {} )"
and  "\<forall>ti l. comploc ( (memory ts)!(coh ts ti l)) l = l"
shows "  OTS ts' t y = {w}"
 using assms
  apply (simp add: step_def )
  apply (subgoal_tac" OTS ts' t y \<noteq> {}")
  prefer 2
  using OTS_notempty_preserved assms(2) assms(8) apply blast
  using assms(2) ld_cobts_lc by blast

(*rule: Transfer rule for conditional observation when OV is singleton*)
corollary  ld_cobv_lc_s : 
 assumes "mem_structured ts"
 and "step t ( Load x r) ts ts'"
 and " \<langle>x  v\<rangle>[y]\<^sub>t  ts = {w}"
and "vbounded ts"
and " x \<noteq> y"
and  " regs ts' t r = v  "
and  "\<forall>ti l. comploc ( (memory ts)!(coh ts ti l)) l = l"
 and " (\<forall> ti addr.[addr]\<^sub>ti ts \<noteq> {} )" 
shows "  [y]\<^sub>t  ts' =  {w}"
 using assms
  apply (simp add: step_def )
  apply (subgoal_tac" [y]\<^sub>t ts' \<noteq> {}")
  prefer 2
  using OV_notempty_preserved assms(2) assms(7) apply blast
  using assms(2) ld_cobv_lc by blast


lemma ld_ots_sub:
  assumes "mem_structured ts"
  and "vbounded ts"
  and  "step t ( Load  y r ) ts ts'"
shows  "OTS ts' t' x   \<subseteq> OTS ts t'  x"
  using assms
 apply (simp add: step_def)
apply (subgoal_tac " memory ts = memory ts' ")
   prefer 2
   apply auto[1]
 apply (subgoal_tac " vrnew ts' t \<ge>  vrnew ts t ")
   prefer 2
   apply (simp add: load_trans_def Let_def OTS_def OTSF_def notoverwritten_def)
   apply (meson assms(3) vrnew_ld_st_sadrr)
  apply (subgoal_tac "  coh ts t x  \<le> coh ts' t x ")
   prefer 2
  apply (metis assms(3) coh_ld_st_dadrr coh_ld_st_sadrr eq_refl)
  apply (simp add: OTS_def OTSF_def notoverwritten_def)
  apply safe
         apply (metis coh_ld_dif dual_order.trans)
  apply (simp add: load_trans_def split:if_splits)
  apply (metis coh_ld_dif dual_order.trans)
  apply (simp add: load_trans_def split:if_splits)
  using max.coboundedI1 apply blast
  apply blast
  apply (metis max.coboundedI2 max.commute)
  apply blast
  apply blast
  apply blast
     apply (metis coh_ld_dif dual_order.trans)
  apply (simp add: load_trans_def split:if_splits)
  apply (metis coh_ld_dif dual_order.trans)
  by (simp add: load_trans_def split:if_splits)


(*rule [LP1]: After reading an  address x the observable values for any location might become less.*)
lemma ld_ov_sub:
  assumes "mem_structured ts"
  and "vbounded ts"
  and  "step t ( Load  y r ) ts ts'"
shows  "[x]\<^sub>t' ts'  \<subseteq> [x]\<^sub>t' ts"
 using assms
  apply (subgoal_tac " OTS ts'  t'  x \<subseteq> OTS ts t'   x ")
   prefer 2
   apply (simp add: ld_ots_sub)
 apply (subgoal_tac " memory ts =memory ts' ")
  prefer 2
  apply (simp add: step_def)
  apply (meson mem_ld)
  apply (simp add: mapval_def)
  by blast

(*rule: Same rule for sinlgeton OV. *)
corollary ld_ov_sub_s:
  assumes "mem_structured ts"
  and "vbounded ts"
  and  "step t ( Load  y r ) ts ts'"
and "  [x]\<^sub>t' ts = {w}"
and "  (\<forall> t addr.  [addr]\<^sub>t ts \<noteq> {} )"
and  "\<forall>ti l. comploc ( (memory ts)!(coh ts ti l)) l = l"
shows  "[x]\<^sub>t' ts'  = [x]\<^sub>t' ts"
  using assms
  by (metis OV_notempty_preserved ld_ov_sub subset_singleton_iff)




lemma ldstep_vpasync:
assumes "step t' ( Load x r) ts ts'"
shows   " vpasync ts' t y =  vpasync ts t y"
  using assms
  apply (simp add: step_def )
  apply (simp add: load_trans_def)
  by auto

lemma ld_opts_ni:
  assumes "mem_structured ts"
  and "vbounded ts"
  and "step t' ( Load x r) ts ts'"
shows " OPTS ts'  y  =  OPTS ts y "
using assms 
  apply (simp add: step_def)
  apply (simp add: OPTS_def)
  apply (subgoal_tac " maxvp ts' y  = maxvp ts y ")
  prefer 2
  apply (simp add :load_trans_def Let_def)
  using select_convs(8) apply auto[1]
 apply (subgoal_tac "memory ts  = memory ts' ")
    prefer 2
  using mem_ld apply blast
  by (metis (no_types, lifting) notoverwritten_def)

lemma ld_oats_ni:
  assumes "mem_structured ts"
  and "vbounded ts"
  and "step t' ( Load x r) ts ts'"
shows " OATS ts' t  y  =  OATS ts t  y "
using assms 
  apply (simp add: step_def)
  apply (simp add: OATS_def)
  apply (subgoal_tac " vpasync ts' t y =  vpasync ts t y")
  prefer 2
  using assms(3) ldstep_vpasync apply blast
  apply (subgoal_tac "memory ts = memory ts'")
   prefer 2
  using mem_ld apply blast
  by (metis (no_types, lifting)  notoverwritten_def)


(*rule [LS2]: After executing a load, the observable persistent values (OPV) for all addresses
remain the same.*)
lemma ld_opv_ni:
  assumes "mem_structured ts"
  and "vbounded ts"
  and "step t' ( Load x r) ts ts'"
shows  "   [y]\<^sub>P ts=  [y]\<^sub>P ts'"
  using assms 
 apply (simp add: step_def)
  apply (subgoal_tac " OPTS ts'  y  =  OPTS ts y")
   prefer 2
  using assms(3) ld_opts_ni apply auto[1]
  apply (subgoal_tac " memory ts = memory ts' ")
   prefer 2
  using mem_ld apply blast
  by auto

(*rule [LS3]: After executing a load, the observable asynchronous values (OAV) for all adresses/threads
remain the same.*)
lemma ld_oav_ni:
  assumes "mem_structured ts"
  and "vbounded ts"
  and "step t' ( Load x r) ts ts'"
shows "   [y]\<^sup>A\<^sub>t ts =  [y]\<^sup>A\<^sub>t ts'"
  using assms
 apply (simp add: step_def)
  apply (subgoal_tac " OATS ts' t  y  =  OATS ts t  y")
   prefer 2
  using assms(3) ld_oats_ni apply auto[1]
  by auto




(*rule [LS7]: The \<lceil>MFENCE y\<rceil> expression still holds after executing a load.*)
lemma ld_m_ord_ni:
 assumes "mem_structured ts"
  and "vbounded ts"
  and  "step t' ( Load  x r ) ts ts'"
  and "\<lceil>MFENCE y\<rceil>\<^sub>t ts"
shows  "\<lceil>MFENCE y\<rceil>\<^sub>t ts'  "
  using assms
  apply (simp add: step_def) apply (case_tac "t \<noteq> t'")
  apply (subgoal_tac " maxcoh ts' t = maxcoh ts t")
  prefer 2
   apply (simp add: load_trans_def Let_def)
   apply force
apply (subgoal_tac " vrnew ts' t = vrnew ts t")
   prefer 2
 apply (simp add: load_trans_def Let_def)
   apply force
  apply (subgoal_tac " last_entry ts' y = last_entry ts y")
  apply simp
  apply (simp add: last_entry_def)
  apply (simp add: last_entry_set_def)
   apply (metis mem_ld)
  apply (subgoal_tac " maxcoh ts' t \<ge> maxcoh ts t")
  prefer 2
   apply (simp add: load_trans_def Let_def)
  apply force
apply (subgoal_tac " vrnew ts' t \<ge> vrnew ts t")
   prefer 2
 apply (simp add: load_trans_def Let_def)
  using assms(3) vrnew_ld_st_sadrr apply presburger  
  apply (simp add: last_entry_def)
  apply (simp add: last_entry_set_def)
  apply safe
  using dual_order.trans le_max_iff_disj mem_ld by presburger


corollary   ld_ots_s : 
  assumes "mem_structured ts"
  and "step t ( Load x r) ts ts'"
 and "vbounded ts"
and "  [y]\<^sub>t ts = {w}"
and "  (\<forall> t addr.  [addr]\<^sub>t ts \<noteq> {} )"
and  "\<forall>ti l. comploc ( (memory ts)!(coh ts ti l)) l = l"
shows "  [y]\<^sub>t ts' = {w}"
  using assms
  apply (subgoal_tac "  [y]\<^sub>t ts' \<noteq> {} ")
   prefer 2
   apply (meson OV_notempty_preserved)
 using ld_ov_sub 
  using ld_nov_ni by fastforce


(*rule [LS5]: If the OV set of x contains only the value of the last write at x, it remains the same after executing a load. *)
corollary  ld_ots_lm_s : 
  assumes "mem_structured ts"
  and "step t ( Load y r) ts ts'"
  and "vbounded ts"
  and  " \<lceil>x\<rceil>\<^sub>t' ts "
  and "  (\<forall> t addr.  OTS ts t addr \<noteq> {} )"
  and  "\<forall>ti l. comploc ( (memory ts)!(coh ts ti l)) l = l"
shows  " \<lceil>x\<rceil>\<^sub>t' ts' "
  using assms
  apply (simp add: step_def)
  apply (subgoal_tac "  OTS ts' t' x  = {last_entry  ts x }")
   prefer 2 
  using  ld_ots_sub[where ts=ts and ts'=ts' and t=t] 
  apply (metis OTS_notempty_preserved assms(2) assms(6) subset_singleton_iff)
 apply(subgoal_tac" last_entry ts x =  last_entry ts' x")
   prefer 2
   apply (simp add: last_entry_def)
   apply (simp add: last_entry_set_def)
  apply (metis mem_ld)
  by force


(*rule: The number of writes on address x with value v remain the same after executing a load.*)
lemma  ld_oc_ni : 
  assumes "mem_structured ts"
  and "step t ( Load y r) ts ts'"
  and "vbounded ts"
shows  " \<lparr>x,v\<rparr>  ts' = \<lparr>x,v\<rparr>  ts "
  using assms
  apply (simp add: step_def)
  apply (simp add:  oc_set_def)
  by fastforce


lemma ld_last_entry_in_ots :
  assumes "mem_structured ts"
and "vbounded ts"
 and "step t ( Load x r) ts ts'"
 and  "\<forall>ti l. comploc ( (memory ts)!(coh ts ti l)) l = l"
shows "  last_entry ts' x \<in> OTS ts' t x "
  using assms
  apply (simp add: step_def)
apply (unfold  OTS_def)
    apply (unfold  OTSF_def)
    apply clarify
    apply (rule_tac x=" last_entry ts' x" in exI)
    apply (rule conjI)
  apply blast
    apply (subgoal_tac"vbounded ts' ")
     prefer 2
  using assms(3) vbounded_preserved apply blast
    apply (subgoal_tac" mem_structured ts' ")
     prefer 2
  using assms(3) mem_structured_preserved apply blast
    apply (intro conjI impI)
        prefer 5
  using last_entry_notoverwritten apply blast
       apply blast
  using last_entry_bounded apply blast
     apply (metis last_entry_loc)
    apply (subgoal_tac"notoverwritten ts' (vrnew ts' t) (last_entry ts' x) x")
     prefer 2
  using last_entry_notoverwritten apply blast
apply (subgoal_tac " coh ts' t x  < length(memory ts') ")
   prefer 2
          apply (subgoal_tac "vbounded ts'")
      prefer 2
    apply blast
          apply (unfold vbounded_def)
          apply meson
         apply (subgoal_tac " 0 \<le>  coh ts' t x")
          prefer 2
   apply blast
 apply (subgoal_tac " \<forall> i. i \<in> last_entry_set ts'  x \<longrightarrow> i \<le> last_entry ts' x")
   prefer 2
   apply (simp add: last_entry_def )
   apply (metis Max.in_idem finite_last_entry_set max_def order_refl)

  apply (subgoal_tac " \<forall> i. 0\<le> i \<and> i < length (memory ts') \<and>  comploc ( memory ts'!i) x = x\<longrightarrow>  i \<in> last_entry_set ts'  x ")
   prefer 2
   apply (simp add: last_entry_set_def)
  using assms(2) assms(3) assms(4) coh_loc_rel_preserved by presburger


lemma ld_last_entry :
  assumes "mem_structured ts"
and "vbounded ts"
 and "step t ( Load x r) ts ts'"
shows  " (last_entry ts x)  =  (last_entry ts' x) " 
  using assms
  apply (simp add: step_def)
 apply (simp add: last_entry_def last_entry_set_def )
  by (metis mem_ld)


(*rule [LP3]: If the OV set for  address x of a thread t' contains only the last write on x and the value of this 
write is v,  and a thread t reads this value on x, then the OV set of x for t after executing the load contains only
the value v. *)
lemma ld_ov_oc_lm_lc :
  assumes "mem_structured ts"
and "vbounded ts"
 and "step t ( Load x r) ts ts'"
and   " \<lparr>x,v\<rparr>  ts = 1"
and  " regs ts' t r = v "
and  " [x]\<^sub>t' ts = {v} "
 and  " \<lceil>x\<rceil>\<^sub>t' ts "
 and  "\<forall>ti l. comploc ( (memory ts)!(coh ts ti l)) l = l"
shows " [x]\<^sub>t  ts' ={v} "
  using assms
 apply (subgoal_tac "  last_entry ts'  x \<in>   OTS ts' t x ")
   prefer 2
  using ld_last_entry_in_ots apply blast
 apply (subgoal_tac "  last_entry ts'  x =   last_entry ts  x ")
   prefer 2
  apply (simp add: ld_last_entry)
  apply (simp add: step_def)
  apply (subgoal_tac " OTS ts' t x  = { last_entry ts  x} ")
  prefer 2
   apply (simp add: OTS_def)
 apply (subgoal_tac " \<nexists> j. ( 0 \<le> j \<and> j < length (memory ts) \<and> compval (memory ts ! j) = v \<and> comploc (memory ts ! j) x = x \<and>( j \<noteq> last_entry ts  x) )")
      prefer 2
    apply simp
    apply (subgoal_tac "comploc (memory ts! (last_entry ts x)) x = x")
     prefer 2
  using last_entry_loc apply blast
     apply (subgoal_tac "compval (memory ts! (last_entry ts x))  = v")
    prefer 2
     apply (simp add: mapval_def)
     apply fastforce
    apply (simp add:  oc_set_def)
  apply (smt (verit, ccfv_threshold) all_not_in_conv card_1_singleton_iff insert_iff last_entry_bounded mem_Collect_eq)
    apply (elim exE)
    apply (subgoal_tac " ind = last_entry ts  x")
     prefer 2
    apply (simp add: load_trans_def Let_def)
    apply (simp add: OTSF_def)
   apply (subgoal_tac " coh ts' t x =  last_entry ts  x")
  prefer 2
    apply (simp add: load_trans_def Let_def)
   apply (subgoal_tac " \<forall> i.( ( 0 \<le> i \<and> i < length (memory ts')) \<and> comploc (memory ts'!i)  x = x) \<longrightarrow> i \<le> last_entry ts'  x")
    prefer 2
    apply (simp add: last_entry_def)
    apply (simp add: last_entry_set_def)
 apply (unfold OTSF_def)
  apply (smt (verit, ccfv_SIG) Collect_cong le_antisym mem_Collect_eq singleton_iff)
  by (metis mem_ld_trans)


(*rule [LS6]: The value of the last write on x remains the same after a load.*)
lemma st_lv_lc:
  assumes "mem_structured ts"
 and "vbounded ts"
and  "step t ( Load x r) ts ts'"
shows " \<lceil>y:v\<rceil> ts' =  \<lceil>y:v\<rceil> ts  "
  using assms
  apply (simp add: step_def)
  apply (subgoal_tac " memory ts = memory ts' ")
   prefer 2
  using mem_ld_trans apply blast
  apply (subgoal_tac " last_entry ts y = last_entry ts' y")
   prefer 2
   apply (simp add: last_entry_def)
   apply (simp add: last_entry_set_def)
  by simp




end