(*Title: View-Based Owicki-Gries Reasoning for Persistent x86-TSO 
  Authors: Eleni Bila, Brijesh Dongol, Ori Lahav, Azalea Raad, John Wickerson*)

theory Sfence_Rules
imports  "../State" "../Language"  "../Assertions"  "../Wellformedness"
begin

(* The lemmas that begin with the  comment "rule [name]: __" are the proof rules regarding the sfence instruction. The
name corresponds to the name of the rule in the paper. 
   All the other lemmas are auxiliary and are used for proving the proof rules.
   Rules that end with "lc" concern mostly local correctness.
   Rules that end with "ni" concern  mostly  non-interference.
   Rules that end with "gen" concern non-interference and local correctness.
   Rules that end with "rel" concern relations between view sets. *) 

lemma sf_ots_ni:
assumes "mem_structured ts"
and "vbounded ts"
 and  "step t'  (sfence) ts ts'"
shows "   (OTS ts' t y = OTS ts t y)"
 using assms 
    apply (simp add: step_def)
   apply (simp add: OTS_def)
   apply (simp add: OTSF_def)
   apply (subgoal_tac "  coh ts'  t' y =  coh ts  t' y")
    prefer 2
  apply (simp add: sfence_trans_def)
  apply auto[1]
  apply (subgoal_tac "vrnew ts' t = vrnew ts t")
  prefer 2
   apply (simp add: sfence_trans_def)
  apply auto[1]
 apply (subgoal_tac "memory ts'  = memory ts ")
  prefer 2
  apply (simp add: sfence_trans_def)
  apply auto[1]
  apply (simp add: sfence_trans_def notoverwritten_def)
  apply (simp add:  vpcommit_nv_def maxvp_nv_def)
  by force

(*rule [SFS1]: After executing an sfence, the observable  values (OV) for all threads/addresses
remain the same*)
lemma sf_ov_ni:
assumes "mem_structured ts"
and "vbounded ts"
 and  "step t'  (sfence) ts ts'"
shows "   [y]\<^sub>t ts=  [y]\<^sub>t ts'"
 using assms 
  apply (simp add: step_def)
  apply (subgoal_tac "OTS ts' t y = OTS ts t y")
  prefer 2
  using assms(3) sf_ots_ni apply auto[1]
  apply (subgoal_tac " memory ts = memory ts'")
   prefer 2
   apply (simp add: sfence_trans_def)
  apply auto[1]
  by (simp add: mapval_def )

lemma mem_sfence:
 assumes "mem_structured ts"
  and "vbounded ts"
and " ts' = sfence_trans t nv mnv ts"
shows "memory ts = memory ts' "
  using assms
  apply (simp add: sfence_trans_def)
  done

lemma opts_oats_sf:
 assumes "mem_structured ts"
  and "vbounded ts"
 and  "step t (sfence) ts ts'"
and "OATS ts t  x  \<noteq> {}"
shows " OPTS ts' x \<subseteq> OATS ts t x "
using assms 
  apply (simp add: step_def  vpcommit_nv_def maxvp_nv_def )
  apply (subgoal_tac "\<forall>k.  k \<in> OPTS ts' x \<longrightarrow> k \<in>  OATS ts t x" )
  prefer 2
apply (simp add: OPTS_def OATS_def )
  apply clarify
  apply safe
   apply (subgoal_tac " k = 0 ")
           prefer 2
  using  mem_sfence
             apply (metis antisym_conv3 mem_structured_def not_less0)
  using gr_implies_not0 apply blast

 apply (subgoal_tac " k = 0 ")
           prefer 2
  using  mem_sfence
            apply (metis antisym_conv3 mem_structured_def not_less0)
           apply (subgoal_tac " (vpasync ts t x) \<le> maxvp (sfence_trans t nv mnv ts) x")
            prefer 2
            apply (simp(no_asm) add: sfence_trans_def)
            apply (metis max.cobounded1 max.coboundedI1)
           apply (smt le_trans mem_sfence notoverwritten_def)
          apply auto[1]
         apply auto[1]
        apply auto[1]
       apply (metis mem_sfence)
      apply auto[1]
     apply (metis mem_sfence)
    apply (metis mem_sfence)
  apply (subgoal_tac " (vpasync ts t (loc (memory (sfence_trans t nv mnv ts) ! k)) \<le> 
(maxvp (sfence_trans t nv mnv ts) (loc (memory (sfence_trans t nv mnv ts) ! k))))")
    prefer 2
    apply (simp(no_asm) add: sfence_trans_def)
    apply (metis max.bounded_iff max.cobounded1)
   apply (smt mem_sfence notoverwritten_def order.trans)
  by auto



(*rule [SFP]: After a thread t executes an sfence, the set of  observable asynchronous
values (OAV) for every address for t, is a subset of its corresponding set of observable values (OV) *)
lemma opv_oav_sf:
 assumes "mem_structured ts"
  and "vbounded ts"
 and  "step t (sfence) ts ts'"
and "OATS ts t  x  \<noteq> {}"
shows "  [x]\<^sub>P ts' \<subseteq>  [x]\<^sup>A\<^sub>t  ts "
  using assms
  apply (subgoal_tac  " OPTS ts' x \<subseteq> OATS ts t x ")
   prefer 2
   apply (simp add: opts_oats_sf)
  apply (subgoal_tac "mapval  ( OATS ts t x)  (memory ts')  = mapval  ( OATS ts t  x)  (memory ts)")
   prefer 2
   apply (simp add: step_def)
   apply auto[1]
  apply (unfold  mapval_def)
  by blast

(*rule:  opv_oav_sf for singleton OV  *)
corollary  opv_oav_sf_s:
assumes "mem_structured ts"
  and "vbounded ts"
 and  "step t (sfence) ts ts'"
and "OATS ts t  x  \<noteq> {}"
  and  "\<forall>ti l. comploc ( (memory ts)!(coh ts ti l)) l = l"
 and " (\<forall>  addr . [addr]\<^sub>P ts \<noteq> {})"
and "   [x]\<^sup>A\<^sub>t ts  =  {w}"
shows "  [x]\<^sub>P ts'  =  {w} "
  using assms
apply (subgoal_tac "   [x]\<^sub>P ts' \<noteq> {}")
   prefer 2
  using OPV_notempty_preserved apply blast 
  apply (simp add: step_def )
  by (metis assms(3) opv_oav_sf subset_singletonD)




lemma sf_opts_sub:
  assumes "mem_structured ts"
  and "vbounded ts"
  and  "step t ( sfence ) ts ts'"
shows  "OPTS ts'    x \<subseteq> OPTS ts   x"
  using assms
  apply (simp add: step_def)
  apply (subgoal_tac " memory ts = memory ts' ")
   prefer 2
  using mem_sfence apply blast
  apply (subgoal_tac " maxvp ts' x \<ge> maxvp ts x ")
   prefer 2
   apply (simp add: sfence_trans_def Let_def maxvp_nv_def  vpcommit_nv_def)
  apply fastforce
  apply (simp add: OPTS_def notoverwritten_def)
  by force

(*rule [SFP]: After executing an  sfence the observable persistent values for any location might become less.*)
lemma sf_opv_sub:
  assumes "mem_structured ts"
  and "vbounded ts"
  and  "step t ( sfence ) ts ts'"
shows  " [x]\<^sub>P ts'   \<subseteq> [x]\<^sub>P ts"
  using assms
  apply (subgoal_tac " OPTS ts'    x \<subseteq> OPTS ts   x ")
   prefer 2
   apply (simp add: sf_opts_sub)
 apply (subgoal_tac " memory ts =memory ts' ")
   prefer 2
   apply (simp add: step_def)
  apply (meson mem_sfence)
  apply (simp add: mapval_def)
  by blast

(*rule [SFS2]: The number of writes on address x with value v remains the same after executing an sfence.*)
lemma  sf_oc_ni : 
  assumes "mem_structured ts"
  and "step t ( sfence) ts ts'"
  and "vbounded ts"
shows  " \<lparr>x,v\<rparr>  ts' =  \<lparr>x,v\<rparr>  ts "
  using assms
  apply (simp add: step_def)
  apply (simp add:  oc_set_def)
  by fastforce

lemma sf_oats_ni:
  assumes "mem_structured ts"
  and "step t ( sfence) ts ts'"
  and "vbounded ts"
  and " t \<noteq> t' "
shows " OATS ts t' x =  OATS ts' t' x"
  using assms
  apply (simp add: step_def)
  apply (subgoal_tac " memory ts  = memory ts'")
  prefer 2
  using mem_sfence apply blast
  apply (simp add: OATS_def)
  apply (simp add: sfence_trans_def vpcommit_nv_def maxvp_nv_def Let_def)
  by (smt (z3) Collect_cong ext_inject notoverwritten_def surjective update_convs(4) update_convs(5) update_convs(8))


(*rule: An sfence that is executed from a thread t does not affect the set of asynchronous
values that any other thread can observe for any location.*)
lemma sf_oav_ni:
  assumes "mem_structured ts"
  and "step t ( sfence) ts ts'"
  and "vbounded ts"
  and " t \<noteq> t' "
shows "  [x]\<^sup>A\<^sub>t'  ts =  [x]\<^sup>A\<^sub>t'  ts'"
  using assms
  apply (subgoal_tac  " OATS ts t' x =  OATS ts' t' x")
  prefer 2
   apply (simp add: sf_oats_ni)
  apply (subgoal_tac  " memory ts =  memory ts'")
   prefer 2
  using step_def apply auto[1]
  by (simp add: mapval_def)


end