(*Title: View-Based Owicki-Gries Reasoning for Persistent x86-TSO 
  Authors: Eleni Bila, Brijesh Dongol, Ori Lahav, Azalea Raad, John Wickerson*)

theory Store_Rules
imports  "../State" "../Language"  "../Assertions"  "../Wellformedness"
begin

(* The lemmas that begin with the  comment "rule [name]: __" are the proof rules regarding the store instruction. The
name corresponds to the name of the rule in the paper. 
   All the other lemmas are auxiliary and are used for proving the proof rules.
   Rules that end with "lc" concern mostly local correctness.
   Rules that end with "ni" concern  mostly  non-interference.
   Rules that end with "gen" concern non-interference and local correctness.
   Rules that end with "rel" concern relations between view sets. *) 


lemma mem_last_l:
  assumes "memory ts' = memory (store_trans ts l v ti)" 
  shows "compval ( memory ts' ! (length( memory ts') -1)) = v"
  using assms
  apply (simp add:  store_trans_def )
  done

lemma mem_last_no:
  assumes  "mem_structured ts"
and  "memory ts' = memory (store_trans ts l v ti)" 
shows "\<forall>i.(  i < length(memory ts') \<and> i \<ge>0) \<longrightarrow> notoverwritten ts' i  ( (length( memory ts') -1)) l"
  using assms
  apply (simp add:  mem_structured_def  notoverwritten_def)
  done


lemma mem_last_coh:
assumes  "mem_structured ts"
and  " ts' = store_trans ts l v ti" 
shows " ( coh ts' ti l =  (length(memory ts') -1)  )"
  using assms
  apply (simp add:  store_trans_def  mem_structured_def  )
  done

lemma coh_less_than_lastentry:
  assumes  "mem_structured  ts"
and  " ts' = store_trans ts l v ti"
shows " \<forall>t. (( 0\<le> t \<and> t < length(memory ts')-1 \<and> 
         l = comploc  (memory ts'!t) l)  \<longrightarrow> coh ts' ti l > t)"
  using assms
  apply(simp add:  mem_structured_def  )
  using assms apply clarify
  apply (case_tac " t=0")
  using assms apply ( simp add:  mem_structured_def )
   apply (subgoal_tac"memory (store_trans ts l v ti) ! 0= Init_Msg " )
    prefer 2
  using assms 
    apply (metis le0 length_greater_0_conv mem_l)
   apply simp
  using assms  mem_last_coh 
   apply auto[1]
  using assms  mem_last_coh 
  by auto

lemma OTS_lastentry:
 assumes  "mem_structured ts"
and  " ts' = store_trans ts x v ti"
shows " OTS ts' ti x = { length(memory ts')-1} "
 using assms 
  apply(simp add:  mem_structured_def   OTS_def OTSF_def store_trans_def notoverwritten_def
  )
  using  mem_last_coh  coh_less_than_lastentry apply simp
  apply (case_tac " 0 \<notin> OTS ts' ti x" )
   prefer 2
  using  assms   apply (simp add: OTS_def OTSF_def notoverwritten_def )
  using assms   apply  (subgoal_tac" ( coh ts' ti x > 0 )" )
  prefer 2
  using assms 
   apply simp
  apply clarsimp
  using  mem_last_coh
  apply(auto simp add: split: Msg.split)
  done



(*rule [SP1]: After executing a store, a thread can only observe the newly written value.*)
lemma st_ov_lc:
  assumes "mem_structured ts"
 and  "step t ( Store  x v) ts ts'"
shows "  [x]\<^sub>t ts' = {v}"
  using assms 
 apply (simp add: step_def)
  apply (subgoal_tac " OTS ts' t x = { length(memory ts')-1}")
   prefer 2
   apply (simp add: OTS_lastentry)
  apply (simp add: mapval_def)
  done



lemma [simp]:
  assumes "mem_structured ts"
  and " i \<in>  OTS ts ti x"
shows " 0 \<le> i  \<and> i < length(memory ts) "
  using assms apply (simp add:  mem_structured_def OTS_def OTSF_def )
  done


lemma [simp]:
  assumes "mem_structured ts"
 and " 0 \<le> i  \<and> i < length(memory ts) "
 and "memory ts ! i = Init_Msg"
shows "i = 0"
  using assms(1)  apply ( unfold  mem_structured_def   )
  apply (subgoal_tac"  length (memory ts) \<noteq>  0 " )
   prefer 2
   apply blast
  using assms(2) assms(3) le_neq_implies_less by blast


lemma coh_st_dt:
assumes "mem_structured ts"
and "vbounded ts"
and " ts' = store_trans ts x v t'"
and "t \<noteq> t'"
shows " coh ts t  x = coh ts' t x"
  using assms
  apply (simp add: store_trans_def)
  done 

lemma vrnew_st_dt:
assumes "mem_structured ts"
and "vbounded ts"
and " ts' = store_trans ts x v t'"
and "t \<noteq> t'"
shows " vrnew ts t   = vrnew ts' t "
  using assms
  apply (simp add: store_trans_def)
  done 

lemma mem_st_dt:
assumes "mem_structured ts"
and "vbounded ts"
and " ts' = store_trans ts x v t'"
and "t \<noteq> t'"
shows " memory ts'   = memory ts  @ [msg x v t'] "
  using assms
  apply (simp add: store_trans_def)
  done

  
lemma st_ots_dt_lc:
  assumes "mem_structured ts"
and "vbounded ts"
 and  "step t' ( Store  x v) ts ts'"
and " t \<noteq> t'"
shows " OTS ts' t x  = OTS ts t x \<union>  { length(memory ts')-1} "
  using assms
  apply (simp add: step_def)
  apply (subgoal_tac "  coh ts t  x = coh ts' t x")
  prefer 2
  using coh_st_dt apply blast
  apply (subgoal_tac "  vrnew ts t  = vrnew  ts' t ")
   prefer 2
  using vrnew_st_dt apply blast
  apply (subgoal_tac "  memory ts'  =  memory ts  @ [msg x v t'] ")
   prefer 2
   apply auto[1] 
  apply (simp add: OTS_def)
  apply (simp add: OTSF_def)
  apply safe
                      apply (simp_all)
  using mem_l apply auto[1]
                    apply (simp add: mem_l)
                   apply (simp add: notoverwritten_def)
                   apply (metis antisym_conv3 le_less less_SucI less_nat_zero_code mem_l store_add)
                  apply (simp add: notoverwritten_def)
                  apply (metis antisym_conv3 le_less less_SucI less_nat_zero_code mem_l store_add)
                 apply (simp add: notoverwritten_def)
  using mem_l apply auto[1]
                apply (simp add: notoverwritten_def)
  using mem_l apply auto[1]
               apply (simp add: notoverwritten_def)
  apply (metis antisym_conv3 le_less less_SucI less_nat_zero_code mem_l store_add)
              apply (simp add: notoverwritten_def)
              apply (simp add: mem_l)
             apply (simp add: notoverwritten_def)
  apply (metis antisym_conv3 le_less less_SucI less_nat_zero_code mem_l store_add)
            apply (simp add: nat_less_le vbounded_def)
           apply (simp add: notoverwritten_def)
          apply (simp add: nth_append)
         apply (simp add: notoverwritten_def)
         apply (simp add: nth_append)
        prefer 3
        apply (simp add: nth_append)
       apply (simp add: notoverwritten_def)
 apply (subgoal_tac " vrnew ts t <  (length (memory ts))")
        prefer 2
        apply(simp add: vbounded_def)
        apply metis
apply (subgoal_tac " vrnew ts t <  Suc(length (memory ts))")
        prefer 2
        apply linarith
       apply simp
       apply (subgoal_tac " \<forall>i.  i \<le> vrnew ts t \<and>  xa < i \<longrightarrow>   loc (memory ts ! i) \<noteq> loc (memory ts ! xa)")
        prefer 2
        apply auto[1]
       apply (simp add: nth_append)
      apply (simp add: notoverwritten_def)
 apply (subgoal_tac " vrnew ts t <  (length (memory ts))")
        prefer 2
        apply(simp add: vbounded_def)
        apply metis
apply (subgoal_tac " vrnew ts t <  Suc(length (memory ts))")
        prefer 2
       apply linarith
      apply (simp add: nth_append)
  apply (simp add: notoverwritten_def)
 apply (subgoal_tac " vrnew ts t <  (length (memory ts))")
        prefer 2
        apply(simp add: vbounded_def)
      apply metis
     apply (simp add: nth_append)
  apply (simp add: notoverwritten_def)
 apply (subgoal_tac " vrnew ts t <  (length (memory ts))")
        prefer 2
        apply(simp add: vbounded_def)
     apply metis
    apply (simp add: nth_append)
  apply (simp add: notoverwritten_def)
 apply (subgoal_tac " vrnew ts t <  (length (memory ts))")
        prefer 2
        apply(simp add: vbounded_def)
     apply metis
   apply (simp add: nth_append)
  apply (simp add: notoverwritten_def)
 apply (subgoal_tac " vrnew ts t <  (length (memory ts))")
        prefer 2
        apply(simp add: vbounded_def)
     apply metis
  by (simp add: nth_append)


(*rule [SP2]: After executing a store on address x, the set of observable values (OV) for address x (for all threads) is augmented with
the newly written value.*)
lemma  st_ov_dt_lc:
  assumes "mem_structured ts"
  and "vbounded ts"
  and "t \<noteq> t'"
  and  "step t' ( Store  x v) ts ts'"
shows "  [x]\<^sub>t ts'  =  [x]\<^sub>t ts  \<union> {v}"
  using assms
  apply (simp add: step_def)
   apply (subgoal_tac  " OTS ts' t x  = OTS ts t x \<union>  { length(memory ts')-1} ")
  prefer 2
  using assms(4) st_ots_dt_lc apply auto[1]
  apply (subgoal_tac " mapval (OTS ts t x) (memory ts @ [msg x v t']) = mapval (OTS ts t x) (memory ts )")
 prefer 2
    apply (subgoal_tac "\<forall> i. i \<in> (OTS ts t x) \<longrightarrow> i < length (memory ts)")
    prefer 2
    apply (simp add: OTS_def)
   apply (unfold mapval_def )
  using mem_l apply auto[1]
  by simp


lemma st_ots_ni:
  assumes "mem_structured ts"
  and "vbounded ts"
 and  "step t' ( Store  x v) ts ts'"
and "x \<noteq> y" 
shows "   (OTS ts' t y = OTS ts t y)"
 using assms 
  apply (simp add: step_def)
   apply (simp add: OTS_def)
   apply (simp add: OTSF_def)
   apply (subgoal_tac "  coh ts'  t' y =  coh ts  t' y")
    prefer 2
  apply (simp add: store_trans_def)
   apply (simp add: store_trans_def)
   apply (simp add:  notoverwritten_def)
   apply safe
                   apply (metis Msg.sel(1) less_antisym nth_append_length)
                     apply (metis dual_order.strict_trans nth_append)
                     apply (metis Msg.sel(1) less_antisym nth_append_length)
                     apply (metis Msg.sel(1) less_SucE nth_append nth_append_length)
  apply (metis antisym_conv3 le_less less_SucI less_nat_zero_code mem_l store_add)
                     apply (metis Msg.distinct(1) less_antisym nth_append_length)
                     apply (metis antisym_conv3 le_less less_SucI less_nat_zero_code mem_l store_add)
                     apply (metis Msg.distinct(1) less_antisym nth_append_length)
                     apply (metis Msg.distinct(1) less_SucE nth_append nth_append_length)
                     apply (metis dual_order.strict_trans nth_append)
                     apply (metis Msg.sel(1) less_antisym nth_append_length)
                     apply (metis antisym_conv3 le_less less_SucI less_nat_zero_code mem_l store_add)
                     apply (metis Msg.sel(1) less_antisym nth_append_length)
                     apply (metis Msg.sel(1) less_SucE nth_append nth_append_length)
                    apply (metis antisym_conv3 le_less less_SucI less_nat_zero_code mem_l store_add)
                   apply linarith
                  apply (metis nth_append)
  using less_Suc_eq apply blast
                 apply (metis nth_append)
  apply (metis less_SucE mem_l nat_less_le order.asym store_add vbounded_def zero_le)
               apply linarith
              apply (metis mem_l nat_less_le not_less_less_Suc_eq order.asym store_add vbounded_def zero_le)
             apply linarith
            apply (metis nth_append)
           apply (metis nth_append)
  using less_Suc_eq apply blast
  apply (metis mem_l nat_less_le not_less_less_Suc_eq order.asym store_add vbounded_def zero_le)
        apply linarith
       apply (metis nth_append)
                      apply (metis mem_l nat_less_le not_less_less_Suc_eq order.asym store_add vbounded_def zero_le)
                      apply (metis Msg.sel(1) less_antisym nth_append_length)
                      apply (metis dual_order.strict_trans nth_append)
                      apply (metis Msg.sel(1) less_antisym nth_append_length)
                      apply (metis Msg.sel(1) less_SucE nth_append nth_append_length)
                      apply (metis antisym_conv3 le_less less_SucI less_nat_zero_code mem_l store_add)
                      apply (metis Msg.distinct(1) less_antisym nth_append_length)
                      apply (metis antisym_conv3 le_less less_SucI less_nat_zero_code mem_l store_add)
                      apply (metis Msg.distinct(1) less_antisym nth_append_length)
                      apply (metis Msg.distinct(1) less_SucE nth_append nth_append_length)
                      apply (metis dual_order.strict_trans nth_append)
                      apply (metis Msg.sel(1) less_antisym nth_append_length)
                      apply (metis antisym_conv3 le_less less_SucI less_nat_zero_code mem_l store_add)
                     apply (metis Msg.sel(1) less_antisym nth_append_length)
                    apply (metis Msg.sel(1) less_SucE nth_append nth_append_length)
                   apply (metis antisym_conv3 le_less less_SucI less_nat_zero_code mem_l store_add)
                  apply linarith
                 apply (metis nth_append)
  using less_SucI apply blast
               apply (metis nth_append)
              apply (metis Msg.sel(1) less_SucE nth_append nth_append_length)
  using less_Suc_eq apply blast
            apply (metis mem_l nat_less_le not_less_less_Suc_eq order.asym store_add vbounded_def zero_le)
  using less_Suc_eq apply blast
          apply (metis nth_append)
         apply (metis nth_append)
        apply linarith
       apply (metis dual_order.strict_trans2 mem_l store_add vbounded_def zero_le)
      apply linarith
     apply (metis nth_append)
    apply (metis dual_order.strict_trans2 mem_l store_add vbounded_def zero_le)
  done

(*rule [WS1]: After executing a store, the observable values (OV) of addresses  different from the newly written address
remain the same*)
lemma st_ov_ni:
  assumes "mem_structured ts"
  and "vbounded ts"
 and  "step t' ( Store  x v) ts ts'"
and "x \<noteq> y" 
shows "   [y]\<^sub>t ts=  [y]\<^sub>t ts'"
 using assms 
  apply (simp add: step_def)
  apply (subgoal_tac "OTS ts' t y = OTS ts t y")
  prefer 2
  using assms(3) st_ots_ni apply auto[1]
  apply (subgoal_tac " mapval (OTS ts t y) (memory ts @ [msg x v t']) = mapval (OTS ts t y) (memory ts )")
    prefer 2
    apply (subgoal_tac "\<forall> i. i \<in> (OTS ts t y) \<longrightarrow> i < length (memory ts)")
     prefer 2
  apply (simp add: OTS_def)
    apply (unfold mapval_def )
    apply (metis (no_types, lifting) image_cong nth_append)
  apply simp
  done


lemma ld_cobts_lc_intro : 
  assumes "mem_structured ts"
  and "t \<noteq> t'"
  and "step t ( Store x v) ts ts'"
 and "vbounded ts"
  and " x \<noteq> y" 
 and " \<forall>k. ( k \<in> OTS ts t'  x  \<longrightarrow> compval (memory ts! k) \<noteq> v )"
 and  "  OTS ts t y =  S "
and " \<forall>ti l. comploc ( (memory ts)!(coh ts ti l)) l = l"

shows "   COBTS x y v t' ts' \<subseteq>  S "
  using assms
  apply (simp add: step_def )
  apply clarify
  apply (subgoal_tac "  { j   | j. j \<in> OTS ts' t'  x \<and> v = compval ((memory ts')!j)} =  { length(memory ts')-1}  ")
  prefer 2
   apply (subgoal_tac "  OTS ts' t' x  = OTS ts t' x \<union>  { length(memory ts')-1} ")
    prefer 2
  using assms(3) st_ots_dt_lc apply auto[1]
   apply (subgoal_tac "  compval (memory ts' ! ( length(memory ts')-1) ) = v ")
    prefer 2
  using mem_last_l apply blast
   apply safe 
    apply (simp )
    apply (subgoal_tac " j \<in> OTS ts t' x \<longrightarrow> compval(memory ts!j) \<noteq> v  ")
     prefer 2
  using assms(6) apply blast
   apply (subgoal_tac " j \<in> OTS ts t' x \<longrightarrow> compval(memory ts'!j) \<noteq> v  ")
     prefer 2
     apply (subgoal_tac " \<forall> j.  j \<in> OTS ts t' x \<longrightarrow> j < length (memory ts)")
      prefer 2
      apply (simp add: OTS_def)
     apply (simp add: nth_append)
    apply simp
   apply simp
   apply (subgoal_tac "  cond_ts ts'  t' x v = { compvrnew ts' t'  ( length(memory ts')-1)  x }")
   prefer 2
   apply (unfold cond_ts_def)
   apply (smt Collect_cong mem_Collect_eq singleton_conv2)
  apply (subgoal_tac "   COBTS x y v t' ts' = OTSF ts' t'  y ( compvrnew ts' t'  ( length(memory ts')-1)  x ) ")
   prefer 2
   apply (simp add: COBTS_def cond_ts_def)
  apply (subgoal_tac " \<forall> j. j \<in>   OTSF ts' t'  y ( compvrnew ts' t'  ( length(memory ts')-1)  x ) \<longrightarrow> j \<in>  OTS ts t y")
   prefer 2
   apply (subgoal_tac " OTS ts t y = OTS ts' t y ")
    prefer 2
  using  st_ots_ni 
  using assms(3) apply blast
  apply (subgoal_tac " \<forall> j. j \<in>   OTSF ts' t'  y ( compvrnew ts' t'  ( length(memory ts')-1)  x ) \<longrightarrow> j \<in>  OTS ts' t y")
  prefer 2
  apply (simp add: OTS_def )
   apply ( unfold OTSF_def )
   apply (case_tac "  coh ts'  t' y >  coh ts' t y")
    apply (subgoal_tac " compvrnew ts'  t' ( length(memory ts) ) x  \<ge> vrnew ts' t ")
     prefer 2
     apply (simp add: compvrnew_def)
     apply (subgoal_tac " length (memory ts)  \<noteq> coh ts t' x ")
      prefer 2
      apply (subgoal_tac "coh ts t' x < length (memory ts)")
       prefer 2
       apply (simp add: vbounded_def)
      apply linarith
     apply simp
     apply (subgoal_tac " vrnew (store_trans ts x v t) t = vrnew ts t")
      prefer 2
      apply (simp add: store_trans_def)
     apply simp
  apply (subgoal_tac "   vrnew ts t
          < (length (memory ts))")
      prefer 2
      apply (simp add: vbounded_def)
     apply linarith
     apply (smt le_eq_less_or_eq le_trans mem_Collect_eq test)
 (***************)
    apply (subgoal_tac "  \<forall> j. ( coh ts'  t'  y \<le> j \<and>  coh ts'  t y > j \<and>   comploc (memory ts' ! j) y =y)  \<longrightarrow>
  \<not> ( notoverwritten ts'(compvrnew ts' t'  (length (memory ts)) x) j y )  ") 
     prefer 2
     apply (simp add: notoverwritten_def compvrnew_def)
     apply safe 
        apply (simp_all)
      apply (rule_tac x=" coh ts'  t y" in exI)
  apply (rule conjI)
       apply (subgoal_tac " vbounded ts' ")
        prefer 2
  using assms(3) vbounded_preserved apply auto[1]
       apply (subgoal_tac " length (memory ts') =  Suc (length (memory ts))")
        prefer 2
        apply simp
       apply (simp add: vbounded_def)
      apply simp
      apply (rule conjI)
       apply (subgoal_tac "   coh ts'  t y =  coh ts  t y")
        prefer 2
        apply (simp add: store_trans_def)
       apply simp
       apply (subgoal_tac "  coh ts t y < length (memory ts)")
        prefer 2
        apply (simp add: vbounded_def)
       apply simp
   apply (subgoal_tac " \<forall>ti l. comploc ( (memory ts')!(coh ts' ti l)) l = l ")
       prefer 2
  using coh_loc_rel_preserved 
  using assms(3) assms(8) apply presburger
      apply simp
  apply (subgoal_tac " (memory ts @ [msg x v t]) ! coh (store_trans ts x v t) t y  \<noteq>
              Init_Msg")
  apply (simp add: mem_structured_def)
       apply meson
(************************************************)
      apply (subgoal_tac " mem_structured ts' ")
       prefer 2
  using store_wfs apply auto[1]
   apply ( simp add: mem_structured_def)
      apply (subgoal_tac "  (memory ts @ [msg x v t]) = memory ts'")
       prefer 2
  using store_add apply presburger
   apply (subgoal_tac " coh (store_trans ts x v t) t y  > 0")
       prefer 2
       apply linarith
      apply (subgoal_tac " coh (store_trans ts x v t) t y<  Suc (length (memory ts))")
       prefer 2
  apply (subgoal_tac " vbounded ts' ")
        prefer 2
  using assms(3) vbounded_preserved 
        apply presburger
       apply(unfold vbounded_def)
       apply (metis length_append_singleton)
      apply blast
(***************************second safe******************)
   apply (rule_tac x=" coh ts'  t y" in exI)
     apply (rule conjI)  (*5*)
   apply (subgoal_tac " vbounded ts' ")
        prefer 2
  using assms(3) vbounded_preserved 
  using assms(4) apply presburger
   apply (subgoal_tac " length (memory ts') =  Suc (length (memory ts))")
        prefer 2
        apply simp
       apply (simp add: vbounded_def)
     apply simp
     apply (rule conjI) (*5*)
       apply (subgoal_tac "    coh ts' t (loc ((memory ts @ [msg x v t]) ! j)) < length( memory ts')")
        prefer 2
       apply (simp add: store_trans_def)
       apply (simp add: less_SucI)
      apply (subgoal_tac "    coh ts' t (loc ((memory ts @ [msg x v t]) ! j)) \<le> length( memory ts)")
  prefer 2
       apply simp
  apply auto[1]
 apply (subgoal_tac " \<forall>ti l. comploc ( (memory ts')!(coh ts' ti l)) l = l ")
       prefer 2
  using coh_loc_rel_preserved 
  using assms(3) assms(8) 
      apply (meson assms(4))
  apply simp
  apply (subgoal_tac " mem_structured ts' ")
       prefer 2
  using store_wfs apply auto[1]
  apply (simp add: mem_structured_def)
 apply (subgoal_tac "  coh (store_trans ts x v t) t
             (loc ((memory ts @ [msg x v t]) ! j)) > 0")
      prefer 2
      apply linarith
  apply (subgoal_tac "  coh (store_trans ts x v t) t
             (loc ((memory ts @ [msg x v t]) ! j)) < Suc (length (memory ts))")
      prefer 2
 apply (subgoal_tac " vbounded ts' ")
        prefer 2
  using assms(3) vbounded_preserved 
  using assms(4) apply presburger
  apply (subgoal_tac " length (memory ts') =  Suc (length (memory ts))")
        prefer 2
       apply simp
      apply (simp add: vbounded_def)
  apply meson
  apply auto[1]
   apply auto[1]
  apply(rule conjI)
   apply (metis (no_types, lifting) less_or_eq_imp_le linorder_neqE_nat)
  apply (subgoal_tac " compvrnew ts'  t' ( length(memory ts) ) x  \<ge> vrnew ts' t ")
     prefer 2
     apply (simp add: compvrnew_def)
     apply (subgoal_tac " length (memory ts)  \<noteq> coh ts t' x ")
      prefer 2
      apply (subgoal_tac "coh ts t' x < length (memory ts)")
       prefer 2
       apply (simp add: vbounded_def)
      apply linarith
     apply simp
     apply (subgoal_tac " vrnew (store_trans ts x v t) t = vrnew ts t")
      prefer 2
      apply (simp add: store_trans_def)
     apply simp
  apply (subgoal_tac "   vrnew ts t
          < (length (memory ts))")
      prefer 2
      apply (simp add: vbounded_def)
   apply linarith
  by (metis (no_types, lifting) le_neq_implies_less test)

lemma opv_opts_rel:
assumes "  v \<notin> [x]\<^sub>t' ts "
shows " \<forall>k. ( k \<in> OTS ts t'  x  \<longrightarrow> compval (memory ts! k) \<noteq> v )"
  using assms
  apply( simp add: mapval_def)
  using inf.strict_coboundedI2 by auto



(*rule [SP5]: Introduction rule for conditional observation*)
lemma ld_cobv_lc_intro : 
  assumes "mem_structured ts"
  and "t \<noteq> t'"
  and "step t ( Store x v) ts ts'"
 and "vbounded ts"
  and " x \<noteq> y"
 and "  v \<notin> [x]\<^sub>t' ts "
 and  "  [y]\<^sub>t ts  =  S "
and " \<forall>ti l. comploc ( (memory ts)!(coh ts ti l)) l = l"
shows "  \<langle>x  v\<rangle>[y]\<^sub>t' ts'  \<subseteq>  S "
 using assms
  apply (simp add: step_def )
  apply (subgoal_tac "  \<forall>k. ( k \<in> OTS ts t'  x  \<longrightarrow> compval (memory ts! k) \<noteq> v )")
   prefer 2
  using opv_opts_rel
   apply blast
 apply (subgoal_tac "  COBTS x y v t' ts' \<subseteq>   OTS ts t y ")
   prefer 2
  using assms(3) ld_cobts_lc_intro apply auto[1]
  apply (subgoal_tac " mapval ( OTS ts t y)  (memory ts')  =mapval ( OTS ts t y)  (memory ts)") 
  prefer 2
   apply (subgoal_tac "\<forall> i. i \<in> (OTS ts t y) \<longrightarrow> i < length (memory ts)")
  prefer 2
    apply (simp add: OTS_def)
  using assms(3) st_ots_ni st_ov_ni apply auto[1]
  apply (unfold mapval_def)
  by (metis image_mono store_add)



lemma [simp]:
 assumes "mem_structured ts"
and "i = 0"
shows "(memory ts ) ! i =  Init_Msg"
  using assms
  apply (simp add: mem_structured_def vbounded_def )
  done 


lemma vp_commit_st:
assumes  "mem_structured ts"
and  " ts' = store_trans ts l v ti'" 
shows " ( vpcommit ts' ti' l =  vpcommit ts ti' l  )" 
 using assms
  apply (simp add:  store_trans_def  )
  done

lemma vp_commit_gen:
assumes  "mem_structured ts"                                  
and  " ts' = store_trans ts l v ti'" 
shows " ( \<forall> ti. ( vpcommit ts' ti l =  vpcommit ts ti l)  )"
 using assms
 apply (simp add: store_trans_def)
  done



lemma OATS_st:
  assumes "mem_structured ts"
and "vbounded ts"
and " i \<in> ( OATS ts ti l)"
and  " ts' = store_trans ts l v t" 
shows " i \<in>  (OATS ts' ti  l)"
 using assms
  apply (simp add: OATS_def )
 apply ( case_tac " i = 0")
   apply (subgoal_tac "(memory ts @ [msg l v t]) ! 0 = Init_Msg")
   prefer 2
   apply (metis mem_structured_def nth_append)
  apply (subgoal_tac " vpasync ts' ti l  = vpasync ts ti l")
   prefer 2
  apply (simp add: store_trans_def)
 apply(subgoal_tac "  vpasync ts ti l  <(length (memory ts)) ")
   prefer 2
   apply(simp add: vbounded_def)
  apply (smt (z3) less_trans mem_l notoverwritten_def order.order_iff_strict)

  apply (subgoal_tac " (memory ts @ [msg l v t]) ! i \<noteq> Init_Msg")
  prefer 2
  apply (metis bot_nat_0.not_eq_extremum less_or_eq_imp_le mem_l mem_structured_def store_add)
 apply (subgoal_tac " vpasync ts' ti l  = vpasync ts ti l")
   prefer 2
   apply (simp add: store_trans_def)
 apply(subgoal_tac "  vpasync ts ti l  <(length (memory ts)) ")
   prefer 2
  apply(simp add: vbounded_def)
  apply simp
  apply (simp add: notoverwritten_def)
  by (smt (z3) dual_order.strict_trans1 less_or_eq_imp_le mem_l not_less store_add)


lemma OPTS_st:
  assumes "mem_structured ts"
and "vbounded ts"
and " i \<in> ( OPTS ts l)"
and  " ts' = store_trans ts l v ti" 
shows " i \<in>  (OPTS ts' l)"
 using assms
  apply (simp add: OPTS_def   )
  apply ( case_tac " i = 0")
   apply (subgoal_tac "(memory ts @ [msg l v ti]) ! 0 = Init_Msg")
   prefer 2
   apply (unfold  mem_structured_def)
   apply simp
   apply (simp add: nth_append)
  apply simp
  apply (simp add: notoverwritten_def)
  apply (subgoal_tac"  maxvp (store_trans ts l v ti) l = maxvp ts l")
   prefer 2
   apply (simp add: store_trans_def)
   apply(subgoal_tac " maxvp ts l < Suc (length (memory ts)) ")
   prefer 2
   apply(simp add: vbounded_def)
  apply (simp add: less_Suc_eq)
   apply(subgoal_tac " maxvp ts l <(length (memory ts)) ")
   prefer 2
   apply(simp add: vbounded_def)
  apply (simp add: nth_append)
   apply (subgoal_tac " (memory ts @ [msg l v ti]) ! i \<noteq> Init_Msg")
  prefer 2
  using mem_l apply auto[1]
   apply simp
   apply (simp add: nth_append)
  apply (simp add: notoverwritten_def)
  apply (subgoal_tac"  maxvp (store_trans ts l v ti) l = maxvp ts l")
   prefer 2
   apply (simp add: store_trans_def)
   apply(subgoal_tac " maxvp ts l < Suc (length (memory ts)) ")
   prefer 2
   apply(simp add: vbounded_def)
  apply (simp add: less_Suc_eq)
   apply(subgoal_tac " maxvp ts l <(length (memory ts)) ")
   prefer 2
   apply(simp add: vbounded_def)
  by (metis le_less_trans nth_append)


lemma [simp]: "mem_structured ts \<Longrightarrow> \<forall>i. 0 < i \<and> i < length(memory ts)  \<longrightarrow>memory ts ! i \<noteq> Init_Msg"
  apply (unfold mem_structured_def)
  by simp

lemma [simp]:  "mem_structured ts \<Longrightarrow> memory ts ! 0 = Init_Msg"
  using mem_structured_def by auto


lemma st_oats_gen:
  assumes "mem_structured ts"
  and "vbounded ts"
  and  "step t ( Store  x v) ts ts'"
shows "  OATS ts' ti x  =  OATS ts ti x \<union> {length (memory ts)}"
  using assms
  apply (simp add: step_def)
  apply (subgoal_tac " \<forall>i. i \<in> OATS ts ti x \<longrightarrow> i \<in> OATS  ts' ti x")
   prefer 2
  using assms OATS_st apply blast
apply (subgoal_tac "length (memory ts) \<in> OATS ts' ti x ")
   prefer 2
  using assms  apply (simp add:  OATS_def )
   apply (unfold notoverwritten_def)
   apply (subgoal_tac "vpasync ts' ti x = vpasync ts ti x")
    prefer 2
    apply (simp add: store_trans_def)
   apply simp
 apply (subgoal_tac" \<nexists>i. i \<in> OATS ts' ti  x \<and> (i \<notin> (OATS ts ti x \<union> {length (memory ts)})) ")
   prefer 2
   apply (simp  add: OATS_def ) 
   apply clarify
   apply (case_tac " i = 0")
  apply (subgoal_tac " length (memory (store_trans ts x v t)) = length(memory ts) +1 ")
     prefer 2
     apply (simp add: store_trans_def)
  apply (subgoal_tac "  memory ts \<noteq> []")
     prefer 2
  using mem_structured_def apply blast
   apply (subgoal_tac "vpasync (store_trans ts x v t) ti x = vpasync ts ti x")
   prefer 2
     apply (simp add: store_trans_def)
 apply simp
    apply (subgoal_tac " (memory ts @ [msg x v t]) ! 0 = Init_Msg ")
     prefer 2
     apply (subgoal_tac " \<forall> i < length(memory ts) . (memory ts @ [msg x v t]) ! i = (memory ts ) ! i")
      prefer 2
      apply (simp add: nth_append)
    apply (subgoal_tac " memory ts ! 0 = Init_Msg")
     prefer 2
  using mem_structured_def apply blast
 apply (subgoal_tac " 0 < length(memory ts)")
      prefer 2
      apply blast
     apply presburger
  apply simp
    apply (simp (no_asm) add:notoverwritten_def)
    apply (metis le0 less_Suc_eq mem_l store_add)
  apply simp
   apply (subgoal_tac " 0 < i \<and> i < length(memory ts)  \<longrightarrow>memory ts ! i \<noteq> Init_Msg")
  prefer 2
    apply simp
 apply (subgoal_tac" vpasync (store_trans ts x v t) ti x = vpasync ts ti x")
   prefer 2
     apply (simp add: store_trans_def)
   apply simp
   apply(intro conjI)
    apply (meson leI le_less_Suc_eq mem_structured_def)
 apply (subgoal_tac " \<forall> i < length(memory ts) . (memory ts @ [msg x v t]) ! i = (memory ts ) ! i")
      prefer 2
    apply (simp add: nth_append)
   apply simp
 apply(safe)
         apply (metis less_antisym)
        apply linarith
       apply (metis less_antisym)
      apply (metis less_SucE)
  using less_antisym apply blast
  apply (smt length_append_singleton lessI less_trans notoverwritten_def store_add)
   apply blast
  by simp 


lemma st_opts_gen:
  assumes "mem_structured ts"
  and "vbounded ts"
  and  "step ti ( Store  x v) ts ts'"
shows "  OPTS ts' x  =  OPTS ts x \<union> {length (memory ts)}"
using assms 
  apply (simp add: step_def)
  apply (subgoal_tac "\<forall> i. i \<in> OPTS ts x \<longrightarrow>  i \<in> OPTS  ts' x")
  prefer 2
  using assms OPTS_st 
  apply blast
  apply (subgoal_tac "length (memory ts) \<in> OPTS ts' x ")
   prefer 2
  using assms  apply (simp add:  OPTS_def )
   apply (unfold notoverwritten_def)
  using  vp_commit_gen 
  using less_le_trans order.asym apply fastforce
  apply (subgoal_tac" \<nexists>i. i \<in> OPTS ts' x \<and> (i \<notin> (OPTS ts x \<union> {length (memory ts)})) ")
   prefer 2
   apply (simp  add: OPTS_def ) 
   apply clarify
   apply (case_tac " i = 0")
    apply (subgoal_tac " length (memory (store_trans ts x v ti)) = length(memory ts) +1 ")
     prefer 2
     apply (simp add: store_trans_def)
    apply (subgoal_tac "  memory ts \<noteq> []")
     prefer 2
  using mem_structured_def apply blast
    apply simp
  apply (subgoal_tac"  maxvp (store_trans ts x v ti) x = maxvp ts x")
   prefer 2
     apply (simp add: store_trans_def)
    apply simp
    apply (subgoal_tac " (memory ts @ [msg x v ti]) ! 0 = Init_Msg ")
     prefer 2
     apply (subgoal_tac " \<forall> i < length(memory ts) . (memory ts @ [msg x v ti]) ! i = (memory ts ) ! i")
      prefer 2
      apply (simp add: nth_append)
    apply (subgoal_tac " memory ts ! 0 = Init_Msg")
     prefer 2
  using mem_structured_def apply blast
     apply (subgoal_tac " 0 < length(memory ts)")
      prefer 2
      apply blast
     apply presburger
  apply simp
    apply (simp (no_asm) add:notoverwritten_def)
    apply (metis le0 less_Suc_eq mem_l store_add)
  apply simp
   apply (subgoal_tac " 0 < i \<and> i < length(memory ts)  \<longrightarrow>memory ts ! i \<noteq> Init_Msg")
  prefer 2
    apply simp
 apply (subgoal_tac"  maxvp (store_trans ts x v ti) x = maxvp ts x")
   prefer 2
     apply (simp add: store_trans_def)
   apply simp
   apply(intro conjI)
    apply (meson leI le_less_Suc_eq mem_structured_def)
 apply (subgoal_tac " \<forall> i < length(memory ts) . (memory ts @ [msg x v ti]) ! i = (memory ts ) ! i")
      prefer 2
    apply (simp add: nth_append)
   apply simp
   apply(safe)
         apply (metis less_antisym)
        apply linarith
       apply (metis less_antisym)
      apply (metis less_SucE)
  using less_antisym apply blast
    apply (smt length_append_singleton lessI less_trans notoverwritten_def store_add)
   apply blast
  by simp

(*rule [SP3]: After executing a store, the sets of asynchronous observable values (OAV) for x (for all threads) are augmented with
the newly written value.*)
lemma st_oav_gen:
  assumes "mem_structured ts"
  and "vbounded ts"
  and  "step ti ( Store  x v) ts ts'"
shows "   [x]\<^sup>A\<^sub>t  ts'  =   [x]\<^sup>A\<^sub>t ts  \<union> {v}"
using assms 
  apply (simp add: step_def )
 apply (subgoal_tac"  OATS ts' t  x  =  OATS ts t x \<union> {length (memory ts)}")
  prefer 2
  using assms(3) st_oats_gen apply blast
  apply (subgoal_tac " mapval  ( OATS ts t x)  (memory ts') = mapval  ( OATS ts t  x)  (memory ts)")
   prefer 2
 apply (subgoal_tac "  length (memory ts') > length (memory ts)") 
  prefer  2
    apply (simp add:  store_trans_def)
   apply (subgoal_tac "\<forall> i. i \<in>  ( OATS ts t x) \<longrightarrow> i < length (memory ts)")
   prefer 2
  apply (simp add:  OATS_def)
   apply (subgoal_tac "\<forall> i. i \<in>  (  OATS ts t x) \<longrightarrow> i < length (memory ts')")
    prefer 2
  using less_trans apply blast
   apply (unfold mapval_def)
   apply (metis (mono_tags, lifting) bot.extremum_strict bot_nat_def image_cong mem_l not_le_imp_less)
  by simp
 

(*rule [SP4]: After executing a store in address x, the set of observable persistent values (OPV) for x is  augmented with
the newly written value.*)
lemma st_opv_gen:
  assumes "mem_structured ts"
  and "vbounded ts"
  and  "step ti ( Store  x v) ts ts'"
shows "   [x]\<^sub>P ts'  =  [x]\<^sub>P ts  \<union> {v}"
using assms 
  apply (simp add: step_def )
 apply (subgoal_tac"  OPTS ts'  x  =  OPTS ts x \<union> {length (memory ts)}")
  prefer 2
  using assms(3) st_opts_gen apply blast
  apply (subgoal_tac " mapval  ( OPTS ts x)  (memory ts') = mapval  ( OPTS ts x)  (memory ts)")
   prefer 2
   apply (subgoal_tac "  length (memory ts') > length (memory ts)") 
  prefer  2
    apply (simp add:  store_trans_def)
   apply (subgoal_tac "\<forall> i. i \<in>  ( OPTS ts x) \<longrightarrow> i < length (memory ts)")
  prefer 2
    apply (simp add:  OPTS_def)
   apply (subgoal_tac "\<forall> i. i \<in>  ( OPTS ts x) \<longrightarrow> i < length (memory ts')")
    prefer 2
  using less_trans apply blast
   apply (unfold mapval_def)
   apply (metis (mono_tags, lifting) bot.extremum_strict bot_nat_def image_cong mem_l not_le_imp_less)
  by simp


lemma st_opts_daddr_ni: 
  assumes "mem_structured ts"
  and "vbounded ts"
 and "x \<noteq> y"
  and  "step ti ( Store  x v) ts ts'"
shows " OPTS ts'  y  =  OPTS ts y "
using assms 
  apply (simp add: step_def)
  apply (simp add: OPTS_def)

  apply (subgoal_tac " maxvp ts' y = maxvp ts y")
  prefer 2
  apply (simp add: store_trans_def)
  apply clarify
  apply safe
  using less_SucE apply fastforce
                      apply (metis Msg.sel(1) less_SucE nth_append nth_append_length)
  using less_antisym apply fastforce
                      apply (metis Msg.sel(1) less_SucE nth_append nth_append_length)
                      apply (simp add: notoverwritten_def)
  apply (metis less_SucI less_Suc_eq_0_disj less_Suc_eq_le mem_l store_add)
  using less_SucE apply fastforce
                      apply (unfold  notoverwritten_def)
  apply (metis length_append_singleton less_SucI less_Suc_eq_0_disj less_nat_zero_code mem_l nat_less_le store_add)
  apply (metis Msg.distinct(1) less_antisym nth_append_length)
  apply (metis Msg.distinct(1) less_SucE nth_append nth_append_length)
                      apply (metis nth_append order.strict_trans)
                     apply (metis Msg.sel(1) less_antisym nth_append_length)
  apply (metis length_append_singleton less_SucI less_Suc_eq_0_disj less_nat_zero_code mem_l nat_less_le store_add)
  apply (metis Msg.sel(1) less_antisym nth_append_length)
  apply (metis Msg.sel(1) less_SucE nth_append nth_append_length)
                 apply (metis length_append_singleton less_SucI less_Suc_eq_0_disj less_Suc_eq_le mem_l store_add)
  using less_Suc_eq apply blast
               apply (simp add: mem_l)
  using less_Suc_eq apply blast
             apply (metis nth_append)
            apply simp
            apply (metis Msg.sel(1) less_SucE nth_append nth_append_length)
  using less_Suc_eq apply blast
          apply (subgoal_tac " xa = 0")
           prefer 2
  using mem_structured_def neq0_conv apply blast
          apply (simp (no_asm))
          apply (metis Msg.sel(1) less_antisym nth_append nth_append_length)
  using less_Suc_eq apply blast
  apply (metis nth_append)
       apply (metis nth_append)
      apply linarith

 apply (subgoal_tac " xa = 0")
           prefer 2
      apply (metis mem_structured_def neq0_conv nth_append)
     apply (simp (no_asm))
  apply (metis Msg.sel(1) less_antisym nth_append nth_append_length)
  using less_Suc_eq apply blast
   apply (metis nth_append)
  apply (simp (no_asm))
  by (metis Msg.sel(1) less_SucE nth_append nth_append_length)


lemma st_oats_daddr_ni: 
  assumes "mem_structured ts"
  and "vbounded ts"
 and "x \<noteq> y"
  and  "step ti ( Store  x v) ts ts'"
shows " OATS ts' t  y  =  OATS ts t  y "
using assms 
  apply (simp add: step_def)
  apply (simp add: OATS_def)
  apply (subgoal_tac " vpasync ts' t  y = vpasync ts t  y")
  prefer 2
  apply (simp add: store_trans_def)
  apply clarify
  apply safe
  using less_antisym apply fastforce
  apply (metis Msg.sel(1) less_SucE nth_append nth_append_length)
  using less_antisym apply fastforce
  apply (metis Msg.sel(1) less_SucE nth_append nth_append_length)
                      apply (simp add: notoverwritten_def)
  apply (metis less_SucI less_Suc_eq_0_disj less_Suc_eq_le mem_l store_add)
  using less_SucE apply fastforce
                      apply (simp(no_asm) add: notoverwritten_def)
                      apply (metis (no_types, lifting) Nat.lessE length_append_singleton less_SucI less_Suc_eq_0_disj mem_l nat_less_le notoverwritten_def store_add)
  using less_SucE apply fastforce
  apply (metis length_append_singleton mem_structured_def neq0_conv store_add store_wfs)
                      apply (metis Msg.distinct(1) less_SucE nth_append nth_append_length)
                     apply (metis Msg.sel(1) less_antisym nth_append_length)
                    apply (simp(no_asm) add: notoverwritten_def)
                    apply (smt Nat.lessE length_append_singleton less_SucI less_Suc_eq_0_disj mem_l nat_less_le notoverwritten_def store_add)
                   apply (metis Msg.sel(1) less_antisym nth_append_length)
                  apply (metis Msg.sel(1) less_SucE nth_append nth_append_length)
                 apply (simp(no_asm) add: notoverwritten_def)
  apply (metis length_append_singleton less_SucI notoverwritten_def nth_append store_add)
  using less_Suc_eq apply blast
               apply (metis nth_append)
  using less_Suc_eq apply blast
  apply (metis nth_append)
            apply (simp(no_asm) add: notoverwritten_def)
  apply (metis Msg.sel(1) less_SucE notoverwritten_def nth_append nth_append_length)
  using less_Suc_eq apply blast
          apply (simp(no_asm) add: notoverwritten_def)
          apply (metis Msg.sel(1) less_SucE notoverwritten_def nth_append nth_append_length)
  using less_Suc_eq apply blast
        apply (metis nth_append)
       apply (metis nth_append)
  using less_Suc_eq apply blast
     apply (simp(no_asm) add: notoverwritten_def)
  apply (metis Msg.sel(1) less_SucE notoverwritten_def nth_append nth_append_length)
  using less_Suc_eq apply blast
   apply (metis nth_append)
     apply (simp(no_asm) add: notoverwritten_def)
  by (metis Msg.sel(1) less_SucE notoverwritten_def nth_append nth_append_length)


(*rule [WS2]: After executing a store, the observable persistent values (OPV) of addresses  different from the newly written address
remain the same*)
lemma st_opv_daddr_ni: 
  assumes "mem_structured ts"
  and "vbounded ts"
  and  "x \<noteq> y"
  and  "step ti ( Store  x v) ts ts'"
shows "   [y]\<^sub>P ts=  [y]\<^sub>P ts'"
 using assms 
  apply (simp add: step_def)
  apply (subgoal_tac " OPTS ts  y = OPTS ts'  y")
  prefer 2
  using assms(4) st_opts_daddr_ni apply blast
 apply (subgoal_tac "mapval  ( OPTS ts y)  (memory ts')  = mapval  ( OPTS ts y)  (memory ts)")
   prefer 2
   apply (subgoal_tac "  length (memory ts') > length (memory ts)") 
  prefer  2
    apply (simp add:  store_trans_def)
   apply (subgoal_tac "\<forall> i. i \<in>  ( OPTS ts y) \<longrightarrow> i < length (memory ts)")
  prefer 2
    apply (simp add:  OPTS_def)
   apply (subgoal_tac "\<forall> i. i \<in>  ( OPTS ts y) \<longrightarrow> i < length (memory ts')")
    prefer 2
     apply (simp add: OPTS_def)
  using Msg.sel(1) apply auto[1]
   apply (unfold  mapval_def)
   apply (metis (no_types, lifting) image_cong le0 mem_l)
  by simp

(*rule [WS3]: After executing a store, the observable asynchronous values (OAV)  of addresses  different from the newly written address
remain the same*)
lemma st_oav_daddr_ni: 
  assumes "mem_structured ts"
  and "vbounded ts"
  and  "x \<noteq> y"
  and  "step ti ( Store  x v) ts ts'"
shows "   [y]\<^sup>A\<^sub>t ts =  [y]\<^sup>A\<^sub>t ts'"
using assms 
  apply (simp add: step_def)
  apply (subgoal_tac " OATS ts t  y = OATS ts' t  y")
  prefer 2
  using assms(4) st_oats_daddr_ni apply auto[1]
 apply (subgoal_tac "mapval  ( OATS ts t y)  (memory ts')  = mapval  ( OATS ts t  y)  (memory ts)")
   prefer 2
   apply (subgoal_tac "  length (memory ts') > length (memory ts)") 
  prefer  2
    apply (simp add:  store_trans_def)
   apply (subgoal_tac "\<forall> i. i \<in>  ( OATS ts t y) \<longrightarrow> i < length (memory ts)")
  prefer 2
    apply (simp add:  OATS_def)
  apply fastforce
   apply (subgoal_tac "\<forall> i. i \<in>  ( OATS ts t y) \<longrightarrow> i < length (memory ts')")
    prefer 2
  using less_trans apply blast
   apply (unfold  mapval_def)
   apply (metis (no_types, lifting) image_cong le0 mem_l)
  by simp


lemma COBS_not_empty:
assumes "mem_structured ts"
and "vbounded ts"
and  "\<forall>ti l. comploc ( (memory ts)!(coh ts ti l)) l = l"
and " cond_ts ts ti x v \<noteq> {}" 
and " \<forall>  nview ti l. OTSF ts ti l nview \<noteq> {}"
shows " ( COBTS x y v ti ts \<noteq> {})"
 using assms apply (simp add:  vbounded_def  step_def )
  apply (simp add: COBTS_def)
  by (metis all_not_in_conv)


lemma COBV_not_empty: 
assumes "mem_structured s"
and "vbounded s"
and  "\<forall>ti l. comploc ( (memory s)!(coh s ti l)) l = l"
and    "v \<in> [x]\<^sub>t s" 
and " \<forall>  nview ti l. OTSF s ti l nview \<noteq> {}"
shows "\<langle>x  v\<rangle>[y]\<^sub>t s  \<noteq> {}"
  using assms
  apply (subgoal_tac " OTS s t x \<noteq> {} ")
   prefer 2
   apply (simp add: OTS_def)  
  apply (subgoal_tac " cond_ts s t x v \<noteq> {}")
  prefer 2
   apply (unfold  cond_ts_def )
  apply (subgoal_tac "  { ta  |ta.
     ta \<in> OTS s t x \<and>
     v = compval (memory s ! ta)} \<noteq>
    {}") 
    prefer 2 
  apply (smt assms(4) disjoint_iff_not_equal image_iff inf_bot_right mapval_def mem_Collect_eq)
  apply (subgoal_tac "  COBTS x y v t s \<noteq> {}")
    prefer 2
    apply (simp add: COBS_not_empty cond_ts_def setcompr_eq_image)
   apply blast
  apply (unfold  mapval_def)
  using COBS_not_empty cond_ts_def by blast
 
(*rule: Introduction rule for conditional observation when OV is singleton *)  
corollary ld_cobv_lc_intro_s:
  assumes "mem_structured ts"
  and "t \<noteq> t'"
  and "step t ( Store x v) ts ts'"
 and "vbounded ts"
  and " x \<noteq> y"
 and "  v \<notin> [x]\<^sub>t' ts "
 and "  v \<in> [x]\<^sub>t' ts' "
 and  "  [y]\<^sub>t ts  =  {w} "
and " \<forall>ti l. comploc ( (memory ts)!(coh ts ti l)) l = l"
and  " \<forall>  nview ti l . OTSF ts ti l nview \<noteq> {}"
shows "  \<langle>x  v\<rangle>[y]\<^sub>t' ts'  =  {w} "
  using assms
  apply (subgoal_tac " \<langle>x  v\<rangle>[y]\<^sub>t' ts' \<subseteq>  [y]\<^sub>t ts")
   prefer 2
  using ld_cobv_lc_intro apply blast
 
  apply (subgoal_tac " \<langle>x  v\<rangle>[y]\<^sub>t' ts' \<noteq>  {}")
   prefer 2
  using COBV_not_empty OTSF_notempty_preserved coh_loc_rel_preserved mem_structured_preserved vbounded_preserved apply presburger
  by (simp add: subset_singleton_iff)



(*rule [WS5]: After executing a store, the last store of all addresses that are different from the newly written address
remain the same*)
lemma st_ots_lm_ni:
  assumes "mem_structured ts"
  and "vbounded ts"
  and  "step t ( Store  x v) ts ts'"
  and "x \<noteq> y"
  and "  \<lceil>y\<rceil>\<^sub>t'  ts "
shows "  \<lceil>y\<rceil>\<^sub>t'  ts'  "
  using assms
 apply (simp add:  last_entry_def )
  apply (subgoal_tac "  OTS ts' t' y = OTS ts t' y")
   prefer 2
   apply (meson st_ots_ni)
  apply (simp add: OTS_def)
  apply (simp add: OTSF_def )
  apply (simp add: step_def)
  apply (unfold last_entry_set_def)
  by (smt (z3) Collect_cong Msg.distinct(1) Msg.sel(1) assms(4) comploc_def length_append_singleton less_Suc_eq mem_l nth_append_length store_add)

(*rule [SP6]: After a thread t executes a store on address x, the observable timestamp of x for t is the last store.*)
lemma st_ots_lm_lc:
  assumes "mem_structured ts"
 and  "step t ( Store  x v) ts ts'"
and " vbounded ts' "
shows " \<lceil>x\<rceil>\<^sub>t ts' "
  using assms
  apply (simp add: step_def)
  apply (subgoal_tac "  OTS ts' t x = { length(memory ts')-1}" )
   prefer 2
  using OTS_lastentry apply auto[1] 
  apply (subgoal_tac " last_entry ts' x = length(memory ts')-1") 
  prefer 2

   apply (subgoal_tac "  comploc ((memory ts') !(length(memory ts')-1)) x = x")
  prefer 2
    apply (simp add: store_trans_def)

   apply (subgoal_tac "( length(memory ts')-1 )\<in> last_entry_set ts' x")
    prefer 2
    apply (simp add: last_entry_set_def)
   apply (subgoal_tac " \<forall>i. (i \<in> last_entry_set ts' x) \<longrightarrow> i \<le> (length(memory ts')-1)")
    prefer 2
    apply (simp add: last_entry_set_def)
   apply (simp add: last_entry_def )
   apply (subgoal_tac" finite (last_entry_set ts' x )")
    prefer 2
  apply (simp add: last_entry_set_def)
  using Max_eqI apply auto[1]
  by simp 

(*rule [SP6]: After a thread t executes  a store on address x  the expression \<lceil>FLUSH y\<rceil> holds for t. *)
lemma st_ord_lc:
assumes  "mem_structured ts"
and "vbounded ts"
 and  "step t ( Store  x v) ts ts'"
shows "\<lceil>FLUSH y\<rceil>\<^sub>t ts' "
  using assms
  apply (simp add: step_def)
  apply (subgoal_tac " maxcoh ts' t  = length (memory ts) ")
  prefer 2
  apply (simp add: store_trans_def)
  apply (subgoal_tac " last_entry ts'  y \<le>  length (memory ts) ")
   prefer 2
   apply (subgoal_tac " last_entry ts'  y \<in>  last_entry_set ts' y ")
    prefer 2
  using assms(3) last_entry_in_last_entry_set store_wfs vbounded_preserved apply blast
   apply (subgoal_tac "   last_entry ts' y < length (memory ts' ) ")
    prefer 2
  using assms(3) last_entry_bounded store_wfs vbounded_preserved apply blast
   apply (simp add: store_trans_def)
  by fastforce 

(*rule [SP9]: After a thread t executes  a store on address x  the expression \<lceil>MFENCE y\<rceil> holds for t. *)
lemma st_m_ord_lc:
assumes  "mem_structured ts"
and "vbounded ts"
 and  "step t ( Store  x v) ts ts'"
shows " \<lceil>MFENCE y\<rceil>\<^sub>t ts' "
  using assms
  apply (simp add: step_def)
  apply (subgoal_tac " maxcoh ts' t  = length (memory ts) ")
  prefer 2
  apply (simp add: store_trans_def)
  apply (subgoal_tac " last_entry ts'  y \<le>  length (memory ts) ")
   prefer 2
   apply (subgoal_tac " last_entry ts'  y \<in>  last_entry_set ts' y ")
    prefer 2
  using assms(3) last_entry_in_last_entry_set store_wfs vbounded_preserved apply blast
   apply (subgoal_tac "   last_entry ts' y < length (memory ts' ) ")
    prefer 2
  using assms(3) last_entry_bounded store_wfs vbounded_preserved apply blast
   apply (simp add: store_trans_def)
  by fastforce 



(*rule [WS6]: If a thread t executes a store on address x,and \<lceil>FLUSH y\<rceil> holds for 
t before executing the store, and x different to y, \<lceil>FLUSH y\<rceil> still holds for t.*)
lemma st_ord_ni:
 assumes "mem_structured ts"
  and "vbounded ts"
  and  "step t' ( Store  x v ) ts ts'"
  and  "\<lceil>FLUSH y \<rceil>\<^sub>t ts "
  and "x \<noteq> y"
shows  "\<lceil>FLUSH y \<rceil>\<^sub>t ts' "
  using assms
  apply (case_tac " t = t'")
  apply (simp add: st_ord_lc)
  apply (simp add: step_def)
   apply (subgoal_tac " last_entry_set ts' y = last_entry_set ts y ")
   prefer 2
   apply (simp add: last_entry_set_def)
  apply (simp add: store_trans_def)
  apply (metis (no_types, lifting) Msg.distinct(1) Msg.sel(1) bot_nat_0.extremum le_imp_less_Suc less_or_eq_imp_le linorder_neqE_nat mem_l not_less_eq nth_append_length store_add)
  apply(simp add: last_entry_def)
  by(simp add: store_trans_def)  


(*rule [WS9]: If a thread t executes  a store on address x, and \<lceil>MFENCE y\<rceil> holds for 
t before executing the store, and x different to y, \<lceil>MFENCE y\<rceil> still holds for t.*)
lemma st_m_ord_ni:
 assumes "mem_structured ts"
  and "vbounded ts"
  and  "step t' ( Store  x v ) ts ts'"
  and  "\<lceil>MFENCE y\<rceil>\<^sub>t ts "
  and "x \<noteq> y"
shows  "\<lceil>MFENCE y\<rceil>\<^sub>t ts' "
  using assms
  apply (case_tac " t = t'")
  apply (simp add: st_m_ord_lc)
  apply (simp add: step_def)
   apply (subgoal_tac " last_entry_set ts' y = last_entry_set ts y ")
   prefer 2
   apply (simp add: last_entry_set_def)
  apply (simp add: store_trans_def)
  apply (metis (no_types, lifting) Msg.distinct(1) Msg.sel(1) bot_nat_0.extremum le_imp_less_Suc less_or_eq_imp_le linorder_neqE_nat mem_l not_less_eq nth_append_length store_add)
  apply(simp add: last_entry_def)
  by(simp add: store_trans_def) 

(*rule [SP7]: The number of writes on x with value v is incremented by 1, after a thread t' executes a store on x with value v. *)
lemma st_oc_lc:
 assumes "mem_structured ts"
  and "vbounded ts"
  and  "step t' ( Store  x v ) ts ts'"
shows " \<lparr>x,v\<rparr>  ts' = \<lparr>x,v\<rparr>   ts +1"
using assms
  apply (simp add: step_def)
 apply (subgoal_tac " oc_set ts'  x v = oc_set ts  x v \<union> {(length (memory ts))}")
  prefer 2
apply (subgoal_tac " \<forall> i. i \<in> oc_set ts' x v   \<longrightarrow> 0 \<le> i \<and> i \<le> (length (memory ts))")
  prefer 2
  apply (subgoal_tac " length (memory ts')  = length (memory ts)+ 1")
   prefer 2
   apply (simp add: store_trans_def)
   apply (simp add: oc_set_def)
  apply (subgoal_tac " \<forall> i. 0 \<le> i \<and> i <  length (memory ts) \<longrightarrow> ( i \<in> oc_set ts' x v \<longrightarrow>  i \<in> oc_set ts x v)")
  prefer 2
  apply (smt (verit, ccfv_SIG) mem_Collect_eq mem_l oc_set_def)
apply (subgoal_tac " length (memory ts)\<in> oc_set ts'  x v ")
  prefer 2
   apply (subgoal_tac "  compval (memory ts' ! (length (memory ts))) = v \<and> comploc (memory ts' ! (length (memory ts))) x = x \<and> 0 \<le> (length (memory ts)) \<and> (length (memory ts)) < length (memory ts') ")
  prefer 2
  apply (simp add: store_trans_def)
  apply (smt (verit, del_insts) mem_Collect_eq oc_set_def)
apply (subgoal_tac " \<nexists> k.  k \<in> oc_set ts' x v \<and> k \<noteq> length (memory ts) \<and> k \<notin> oc_set ts x v ")
  prefer 2
  apply (meson le_imp_less_or_eq)
  apply (subgoal_tac " \<forall> i. i \<in> oc_set ts x v   \<longrightarrow>  i \<in> oc_set ts' x v")
  prefer 2
  apply (smt (z3) less_trans mem_Collect_eq mem_l oc_set_def)
  apply blast
  apply (subgoal_tac " length (memory ts) \<notin>  oc_set ts x v")
  prefer 2
  apply (simp add: oc_set_def)
 apply (subgoal_tac " card ( oc_set ts' x v) = card (oc_set ts x v) + card {length (memory ts)}")
  prefer 2
  apply (subgoal_tac " finite( oc_set ts' x v)  \<and>  finite( oc_set ts x v)   ")
  prefer 2
  apply (simp add: oc_set_def)
  apply simp
  by simp



(*rule [WS7]: The number of writes on x with value v remains the same after a thread t' executes a store on x with value different from v. *)
lemma st_noc_lc:
 assumes "mem_structured ts"
  and "vbounded ts"
  and  "step t' ( Store  x v ) ts ts'"
and "v \<noteq> v'"
shows " \<lparr>y,v'\<rparr>  ts' =  \<lparr>y,v'\<rparr>  ts"
 using assms
  apply (simp add: step_def)
  apply (subgoal_tac " \<forall> j. 0 \<le> j \<and> j < length (memory ts) \<longrightarrow> ( (memory ts' ! j) = (memory ts ! j))")
  prefer 2
    apply (simp add: store_trans_def)
  apply (metis bot.extremum bot_nat_def mem_l store_add)

  apply (simp add: store_trans_def)
  apply (simp add:  oc_set_def)
  by (smt (verit, best) Collect_cong Msg.distinct(1) Msg.sel(2) less_Suc_eq nth_append_length)


(*rule [WS8]: The value of the last write  on x remains the same after a thread t' executes a store on an address different from x.*)
lemma st_lv_ni:
  assumes "mem_structured ts"
 and "vbounded ts"
and  "step t' ( Store  y v ) ts ts'"
and "x \<noteq> y"
shows " \<lceil>x:v\<rceil> ts' =\<lceil>x:v\<rceil> ts  "
  using assms
  apply (simp add: step_def)
  apply (subgoal_tac "  last_entry ts'  x = last_entry ts  x")
   prefer 2 
  apply (simp add: last_entry_def)
   apply (simp add: last_entry_set_def)
   apply (subgoal_tac"( \<forall>i.(i\<ge>0\<and>i<length(memory ts)) \<longrightarrow> ( memory ts)!i =( memory ts')!i)")
    prefer 2
  using mem_l apply blast   
   apply simp   
  apply (metis (no_types, lifting) Collect_cong Msg.discI Msg.sel(1) less_Suc_eq not_less_less_Suc_eq nth_append_length)
  by (metis last_entry_bounded mem_l store_add zero_le)


(*rule [SP8]: The value of the last write  on x becomes v after a thread t' executes a store on x with value v.*)
lemma st_lv_lc:
  assumes "mem_structured ts"
 and "vbounded ts"
and  "step t' ( Store  x v ) ts ts'"
shows " \<lceil>x:v\<rceil> ts' "
  using assms
  apply (simp add: step_def)
  apply (subgoal_tac " last_entry ts' x = length(memory ts')-1") 
   prefer 2
  apply (subgoal_tac "  comploc ((memory ts') !(length(memory ts')-1)) x = x")
  prefer 2
    apply (simp add: store_trans_def)
   apply (subgoal_tac "( length(memory ts')-1 )\<in> last_entry_set ts' x")
    prefer 2
    apply (simp add: last_entry_set_def)
   apply (subgoal_tac " \<forall>i. (i \<in> last_entry_set ts' x) \<longrightarrow> i \<le> (length(memory ts')-1)")
    prefer 2
    apply (simp add: last_entry_set_def)
   apply (simp add: last_entry_def )
   apply (subgoal_tac" finite (last_entry_set ts' x )")
    prefer 2
  apply (simp add: last_entry_set_def)
  using Max_eqI apply auto[1]
  by (simp add: store_trans_def)


lemma st_last_entry_init_message_ni:
  assumes "mem_structured ts"
 and "vbounded ts"
and "memory ts ! last_entry ts lx \<noteq> Init_Msg "
and  "step t' ( Store  x v ) ts ts'"
and " lx \<noteq> x"
shows "memory ts' ! last_entry ts' lx \<noteq> Init_Msg "
  using assms
  apply (simp add: step_def)
  apply (subgoal_tac " last_entry ts lx = last_entry ts' lx ")
   prefer 2 
  apply (simp add: last_entry_def)
   apply (simp add: last_entry_set_def)
   apply (subgoal_tac"( \<forall>i.(i\<ge>0\<and>i<length(memory ts)) \<longrightarrow> ( memory ts)!i =( memory ts')!i)")
    prefer 2
  using mem_l apply blast   
   apply simp   
  apply (metis (no_types, lifting) Collect_cong Msg.discI Msg.sel(1) less_Suc_eq nth_append_length)
  by (metis bot_nat_0.extremum last_entry_bounded mem_l store_add)


lemma st_last_entry_val_ni:
  assumes "mem_structured ts"
 and "vbounded ts"
and "val (memory ts ! last_entry ts lx) = y"
and  "step t' ( Store  x v ) ts ts'"
and " lx \<noteq> x"
shows "val (memory ts' ! last_entry ts' lx) = y"
  using assms
  apply (simp add: step_def)
  apply (subgoal_tac " last_entry ts lx = last_entry ts' lx ")
   prefer 2 
  apply (simp add: last_entry_def)
   apply (simp add: last_entry_set_def)
   apply (subgoal_tac"( \<forall>i.(i\<ge>0\<and>i<length(memory ts)) \<longrightarrow> ( memory ts)!i =( memory ts')!i)")
    prefer 2
  using mem_l apply blast   
   apply simp
   apply (metis (no_types, lifting) Collect_cong Msg.discI Msg.sel(1) less_Suc_eq nth_append_length)
  by (metis bot_nat_0.extremum last_entry_bounded mem_l store_add)



end
 

